import copy
import datetime as dt

import habanero
import pytest
import requests_mock

from regolith.runcontrol import DEFAULT_RC
from regolith.tools import (
    awards_grants_honors,
    collect_appts,
    collection_str,
    compound_dict,
    compound_list,
    create_repo,
    date_to_rfc822,
    dereference_institution,
    filter_employment_for_advisees,
    filter_presentations,
    filter_publications,
    fragment_retrieval,
    fuzzy_retrieval,
    get_appointments,
    get_formatted_crossref_reference,
    get_id_from_name,
    get_person_contact,
    get_tags,
    get_target_repo_info,
    get_target_token,
    get_uuid,
    grant_burn,
    group,
    group_member_employment_start_end,
    group_member_ids,
    is_fully_appointed,
    key_value_pair_filter,
    latex_safe,
    merge_collections_all,
    merge_collections_intersect,
    merge_collections_superior,
    month_and_year,
    number_suffix,
    remove_duplicate_docs,
    search_collection,
    update_schemas,
    validate_meeting,
)

PEOPLE_COLL = [
    {
        "_id": "m1",
        "name": "member1",
        "education": [
            {"group": "bg", "institution": "columbiau", "degree": "PhD", "department": "apam", "begin_year": 2016}
        ],
        "employment": [
            {
                "begin_year": 2020,
                "begin_month": 1,
                "organization": "columbiau",
                "position": "Undergraduate Researcher",
                "advisor": "sbillinge",
                "status": "undergrad",
            }
        ],
        "funding": [{"name": "Omega Laser User's Group Travel Award", "value": 1100, "year": 2013}],
        "service": [
            {
                "name": "International Steering Committee",
                "role": "chair",
                "type": "profession",
                "year": 2020,
                "month": 3,
                "notes": ["something"],
            }
        ],
    },
    {
        "_id": "nm1",
        "name": "non-member1",
        "education": [{"institution": "columbiau", "degree": "PhD", "department": "apam", "begin_year": 2016}],
        "employment": [
            {
                "begin_year": 2020,
                "begin_month": 1,
                "organization": "columbiau",
                "position": "Undergraduate Researcher",
                "advisor": "sbillinge",
                "status": "undergrad",
            }
        ],
    },
    {
        "_id": "m2",
        "name": "member2",
        "education": [{"institution": "columbiau", "degree": "PhD", "department": "apam", "begin_year": 2016}],
        "employment": [
            {
                "begin_year": 2020,
                "begin_month": 1,
                "group": "bg",
                "organization": "columbiau",
                "position": "Undergraduate Researcher",
                "advisor": "sbillinge",
                "status": "undergrad",
            }
        ],
    },
]

CONTACTS_COLL = [{"_id": "c1", "name": "contact1", "institution": "columbiau"}]


@pytest.mark.parametrize(
    "input, expected",
    [
        (
            ["m1", PEOPLE_COLL, CONTACTS_COLL],
            {
                "_id": "m1",
                "funding": [{"name": "Omega Laser User's Group Travel Award", "value": 1100, "year": 2013}],
                "name": "member1",
                "service": [
                    {
                        "month": 3,
                        "name": "International Steering Committee",
                        "notes": ["something"],
                        "role": "chair",
                        "type": "profession",
                        "year": 2020,
                    }
                ],
                "education": [
                    {
                        "group": "bg",
                        "institution": "columbiau",
                        "degree": "PhD",
                        "department": "apam",
                        "begin_year": 2016,
                    }
                ],
                "employment": [
                    {
                        "begin_year": 2020,
                        "begin_month": 1,
                        "organization": "columbiau",
                        "position": "Undergraduate Researcher",
                        "advisor": "sbillinge",
                        "status": "undergrad",
                    }
                ],
            },
        ),
        (["c1", PEOPLE_COLL, CONTACTS_COLL], {"_id": "c1", "name": "contact1", "institution": "columbiau"}),
        (["bad1", PEOPLE_COLL, CONTACTS_COLL], None),
    ],
)
def test_get_person_contact(input, expected):
    actual = get_person_contact(input[0], input[1], input[2])
    assert actual == expected


CITATIONS = [
    {
        "_id": "paper",
        "author": ["m1", "cleese"],
        "ackno": "thanks",
        "grant": "fwp, dmref",
        "month": "apr",
        "year": "2021",
    },
    {"_id": "paper2", "author": ["m1", "palin"], "ackno": "thanks", "grant": "fwp2", "year": "2020"},
    {
        "_id": "paper3",
        "author": ["m1", "jones"],
        "ackno": "thanks",
        "grant": "fwp2",
        "month": "jun",
        "year": "2020",
    },
]


@pytest.mark.parametrize(
    "args, kwargs, expected",
    [
        (
            [CITATIONS, set(["m1"])],
            {},
            [
                {
                    "_id": "paper3",
                    "author": ["\\textbf{m1}", "jones"],
                    "ackno": "thanks",
                    "grant": "fwp2",
                    "month": "jun",
                    "year": "2020",
                },
                {
                    "_id": "paper2",
                    "author": ["\\textbf{m1}", "palin"],
                    "ackno": "thanks",
                    "grant": "fwp2",
                    "year": "2020",
                },
                {
                    "_id": "paper",
                    "ackno": "thanks",
                    "author": ["\\textbf{m1}", "cleese"],
                    "grant": "fwp, dmref",
                    "month": "apr",
                    "year": "2021",
                },
            ],
        ),
        (
            [CITATIONS, set(["m1"])],
            {"bold": False, "ackno": True},
            [
                {
                    "_id": "paper3",
                    "author": ["m1", "jones"],
                    "ackno": "thanks",
                    "grant": "fwp2",
                    "month": "jun",
                    "note": "\\newline\\newline\\noindent Acknowledgement:\\newline\\noindent thanks"
                    "\\newline\\newline\\noindent ",
                    "year": "2020",
                },
                {
                    "_id": "paper2",
                    "author": ["m1", "palin"],
                    "ackno": "thanks",
                    "grant": "fwp2",
                    "note": "\\newline\\newline\\noindent Acknowledgement:\\newline\\noindent thanks"
                    "\\newline\\newline\\noindent ",
                    "year": "2020",
                },
                {
                    "_id": "paper",
                    "ackno": "thanks",
                    "author": ["m1", "cleese"],
                    "grant": "fwp, dmref",
                    "month": "apr",
                    "note": "\\newline\\newline\\noindent Acknowledgement:\\newline\\noindent thanks"
                    "\\newline\\newline\\noindent ",
                    "year": "2021",
                },
            ],
        ),
        (
            [CITATIONS, set(["m1"])],
            {"bold": False},
            [
                {
                    "_id": "paper3",
                    "author": ["m1", "jones"],
                    "ackno": "thanks",
                    "grant": "fwp2",
                    "month": "jun",
                    "year": "2020",
                },
                {"_id": "paper2", "author": ["m1", "palin"], "ackno": "thanks", "grant": "fwp2", "year": "2020"},
                {
                    "_id": "paper",
                    "ackno": "thanks",
                    "author": ["m1", "cleese"],
                    "grant": "fwp, dmref",
                    "month": "apr",
                    "year": "2021",
                },
            ],
        ),
        (
            [CITATIONS, set(["m1"])],
            {"bold": False, "grants": "fwp2"},
            [
                {
                    "_id": "paper3",
                    "author": ["m1", "jones"],
                    "ackno": "thanks",
                    "grant": "fwp2",
                    "month": "jun",
                    "year": "2020",
                },
                {"_id": "paper2", "author": ["m1", "palin"], "ackno": "thanks", "grant": "fwp2", "year": "2020"},
            ],
        ),
        (
            [CITATIONS, set(["m1"])],
            {"bold": False, "grants": ["fwp2", "dmref"]},
            [
                {
                    "_id": "paper3",
                    "author": ["m1", "jones"],
                    "ackno": "thanks",
                    "grant": "fwp2",
                    "month": "jun",
                    "year": "2020",
                },
                {"_id": "paper2", "author": ["m1", "palin"], "ackno": "thanks", "grant": "fwp2", "year": "2020"},
                {
                    "_id": "paper",
                    "ackno": "thanks",
                    "author": ["m1", "cleese"],
                    "grant": "fwp, dmref",
                    "month": "apr",
                    "year": "2021",
                },
            ],
        ),
        (
            [CITATIONS, set(["m1"])],
            {"bold": False, "since": dt.date(2021, 1, 1)},
            [
                {
                    "_id": "paper",
                    "ackno": "thanks",
                    "author": ["m1", "cleese"],
                    "grant": "fwp, dmref",
                    "month": "apr",
                    "year": "2021",
                }
            ],
        ),
        (
            [CITATIONS, set(["m1"])],
            {"bold": False, "since": dt.date(2020, 5, 1), "before": dt.date(2021, 1, 1)},
            [
                {
                    "_id": "paper3",
                    "author": ["m1", "jones"],
                    "ackno": "thanks",
                    "grant": "fwp2",
                    "month": "jun",
                    "year": "2020",
                },
                {"_id": "paper2", "author": ["m1", "palin"], "ackno": "thanks", "grant": "fwp2", "year": "2020"},
            ],
        ),
    ],
)
def test_filter_publications(args, kwargs, expected):
    actual = filter_publications(*args, **kwargs)
    assert actual == expected


def test_author_publications():
    citations = [{"author": ["CJ", "SJLB"]}, {"editor": "SJLB"}]
    filter_publications(citations, {"SJLB"})


def test_fuzzy_retrieval():
    person = {
        "_id": "scopatz",
        "aka": [
            "Scopatz",
            "Scopatz, A",
            "Scopatz, A.",
            "Scopatz, A M",
            "Anthony Michael Scopatz",
        ],
        "name": "Anthony Scopatz",
    }
    assert fuzzy_retrieval([person], ["aka", "name", "_id"], "scopatz") == person
    assert fuzzy_retrieval([person], ["aka", "name", "_id"], "scopatz, a") is None
    assert (
        fuzzy_retrieval(
            [person],
            ["aka", "name", "_id"],
            "scopatz, a",
            case_sensitive=False,
        )
        == person
    )


def test_get_formatted_crossref_reference(monkeypatch):
    def mockreturn(*args, **kwargs):
        mock_article = {
            "message": {
                "author": [{"given": "SJL", "family": "Billinge"}],
                "short-container-title": ["J. Great Results"],
                "volume": 10,
                "title": ["Whamo"],
                "page": "231-233",
                "issued": {"date-parts": [[1971, 8, 20]]},
            }
        }
        return mock_article

    monkeypatch.setattr(habanero.Crossref, "works", mockreturn)
    expected = ("Whamo, SJL Billinge, J. Great Results, v. 10, pp. 231-233, (1971).", dt.date(1971, 8, 20))
    actual = get_formatted_crossref_reference("test")
    assert actual == expected


@pytest.mark.parametrize(
    "input,expected",
    [
        (0, "th"),
        (1, "st"),
        (2, "nd"),
        (3, "rd"),
        (4, "th"),
        (10, "th"),
        (13, "th"),
        (33, "rd"),
        (None, ""),
        ("0", ""),
    ],
)
def test_number_suffix(input, expected):
    assert number_suffix(input) == expected


@pytest.mark.parametrize(
    "input,expected",
    [
        (
            (
                [
                    {
                        "_id": "proposal1",
                        "title": "European swallow",
                        "author": "king arthur",
                    }
                ],
                [{"_id": "grant1", "linked_to": "proposal1", "amount": "100 mph"}],
            ),
            [
                {
                    "_id": "grant1",
                    "title": "European swallow",
                    "author": "king arthur",
                    "linked_to": "proposal1",
                    "amount": "100 mph",
                }
            ],
        ),
        (
            (
                [
                    {
                        "_id": "proposal1",
                        "title": "European swallow",
                        "author": "king arthur",
                    },
                    {
                        "_id": "proposal2",
                        "title": "African swallow",
                        "author": "king arthur",
                    },
                ],
                [{"_id": "grant1", "linked_to": "proposal1", "amount": "100 mph"}],
            ),
            [
                {
                    "_id": "proposal2",
                    "title": "African swallow",
                    "author": "king arthur",
                },
                {
                    "_id": "grant1",
                    "title": "European swallow",
                    "author": "king arthur",
                    "linked_to": "proposal1",
                    "amount": "100 mph",
                },
            ],
        ),
        (
            (
                [
                    {
                        "_id": "proposal1",
                        "title": "European swallow",
                        "author": "king arthur",
                        "amount": "50 mph",
                    },
                    {
                        "_id": "proposal2",
                        "title": "African swallow",
                        "author": "king arthur",
                    },
                ],
                [{"_id": "grant1", "linked_to": "proposal1", "amount": "100 mph"}],
            ),
            [
                {
                    "_id": "proposal2",
                    "title": "African swallow",
                    "author": "king arthur",
                },
                {
                    "_id": "grant1",
                    "title": "European swallow",
                    "author": "king arthur",
                    "linked_to": "proposal1",
                    "amount": "100 mph",
                },
            ],
        ),
        (
            (
                [
                    {
                        "_id": "proposal1",
                        "title": "European swallow",
                        "author": "king arthur",
                        "amount": "50 mph",
                    },
                    {
                        "_id": "proposal2",
                        "title": "African swallow",
                        "author": "king arthur",
                    },
                ],
                [
                    {"_id": "grant1", "linked_to": "proposal1", "amount": "100 mph"},
                    {
                        "_id": "grant2",
                        "title": "African swallow",
                        "author": "king arthur",
                    },
                ],
            ),
            [
                {
                    "_id": "proposal2",
                    "title": "African swallow",
                    "author": "king arthur",
                },
                {
                    "_id": "grant1",
                    "title": "European swallow",
                    "author": "king arthur",
                    "linked_to": "proposal1",
                    "amount": "100 mph",
                },
                {
                    "_id": "grant2",
                    "title": "African swallow",
                    "author": "king arthur",
                },
            ],
        ),
        (
            (
                [
                    {
                        "_id": "proposal1",
                        "title": "European swallow",
                        "author": "king arthur",
                        "amount": "50 mph",
                    },
                    {
                        "_id": "proposal2",
                        "title": "African swallow",
                        "author": "king arthur",
                    },
                ],
                [
                    {"_id": "grant1", "linked_to": "proposal1", "amount": "100 mph"},
                    {
                        "_id": "grant2",
                        "title": "African swallow",
                        "author": "king arthur",
                        "linked_to": "proposal2",
                    },
                ],
            ),
            [
                {
                    "_id": "grant1",
                    "title": "European swallow",
                    "author": "king arthur",
                    "linked_to": "proposal1",
                    "amount": "100 mph",
                },
                {
                    "_id": "grant2",
                    "title": "African swallow",
                    "author": "king arthur",
                    "linked_to": "proposal2",
                },
            ],
        ),
        (
            (
                [
                    {
                        "_id": "proposal1",
                        "title": "European swallow",
                        "author": "king arthur",
                        "amount": "50 mph",
                    },
                ],
                [
                    {"_id": "grant1", "linked_to": "proposal1", "amount": "100 mph"},
                    {
                        "_id": "grant2",
                        "title": "African swallow",
                        "author": "king arthur",
                    },
                ],
            ),
            [
                {
                    "_id": "grant1",
                    "title": "European swallow",
                    "author": "king arthur",
                    "linked_to": "proposal1",
                    "amount": "100 mph",
                },
                {
                    "_id": "grant2",
                    "title": "African swallow",
                    "author": "king arthur",
                },
            ],
        ),
    ],
)
def test_merge_collections_all(input, expected):
    a = input[0]
    b = input[1]
    target_id = "linked_to"
    assert merge_collections_all(a, b, target_id) == expected


@pytest.mark.parametrize(
    "input,expected",
    [
        (
            (
                [
                    {
                        "_id": "proposal1",
                        "title": "European swallow",
                        "author": "king arthur",
                    }
                ],
                [{"_id": "grant1", "linked_to": "proposal1", "amount": "100 mph"}],
            ),
            [
                {
                    "_id": "grant1",
                    "title": "European swallow",
                    "author": "king arthur",
                    "linked_to": "proposal1",
                    "amount": "100 mph",
                }
            ],
        ),
        (
            (
                [
                    {
                        "_id": "proposal1",
                        "title": "European swallow",
                        "author": "king arthur",
                    },
                    {
                        "_id": "proposal2",
                        "title": "African swallow",
                        "author": "king arthur",
                    },
                ],
                [{"_id": "grant1", "linked_to": "proposal1", "amount": "100 mph"}],
            ),
            [
                {
                    "_id": "grant1",
                    "title": "European swallow",
                    "author": "king arthur",
                    "linked_to": "proposal1",
                    "amount": "100 mph",
                },
            ],
        ),
        (
            (
                [
                    {
                        "_id": "proposal1",
                        "title": "European swallow",
                        "author": "king arthur",
                        "amount": "50 mph",
                    },
                    {
                        "_id": "proposal2",
                        "title": "African swallow",
                        "author": "king arthur",
                    },
                ],
                [{"_id": "grant1", "linked_to": "proposal1", "amount": "100 mph"}],
            ),
            [
                {
                    "_id": "grant1",
                    "title": "European swallow",
                    "author": "king arthur",
                    "linked_to": "proposal1",
                    "amount": "100 mph",
                },
            ],
        ),
        (
            (
                [
                    {
                        "_id": "proposal1",
                        "title": "European swallow",
                        "author": "king arthur",
                        "amount": "50 mph",
                    },
                    {
                        "_id": "proposal2",
                        "title": "African swallow",
                        "author": "king arthur",
                    },
                ],
                [
                    {"_id": "grant1", "linked_to": "proposal1", "amount": "100 mph"},
                    {
                        "_id": "grant2",
                        "title": "African swallow",
                        "author": "king arthur",
                    },
                ],
            ),
            [
                {
                    "_id": "grant1",
                    "title": "European swallow",
                    "author": "king arthur",
                    "linked_to": "proposal1",
                    "amount": "100 mph",
                },
                {
                    "_id": "grant2",
                    "title": "African swallow",
                    "author": "king arthur",
                },
            ],
        ),
        (
            (
                [
                    {
                        "_id": "proposal1",
                        "title": "European swallow",
                        "author": "king arthur",
                        "amount": "50 mph",
                    },
                    {
                        "_id": "proposal2",
                        "title": "African swallow",
                        "author": "king arthur",
                    },
                ],
                [
                    {"_id": "grant1", "linked_to": "proposal1", "amount": "100 mph"},
                    {
                        "_id": "grant2",
                        "title": "African swallow",
                        "author": "king arthur",
                        "linked_to": "proposal2",
                    },
                ],
            ),
            [
                {
                    "_id": "grant1",
                    "title": "European swallow",
                    "author": "king arthur",
                    "linked_to": "proposal1",
                    "amount": "100 mph",
                },
                {
                    "_id": "grant2",
                    "title": "African swallow",
                    "author": "king arthur",
                    "linked_to": "proposal2",
                },
            ],
        ),
        (
            (
                [
                    {
                        "_id": "proposal1",
                        "title": "European swallow",
                        "author": "king arthur",
                        "amount": "50 mph",
                    },
                ],
                [
                    {"_id": "grant1", "linked_to": "proposal1", "amount": "100 mph"},
                    {
                        "_id": "grant2",
                        "title": "African swallow",
                        "author": "king arthur",
                    },
                ],
            ),
            [
                {
                    "_id": "grant1",
                    "title": "European swallow",
                    "author": "king arthur",
                    "linked_to": "proposal1",
                    "amount": "100 mph",
                },
                {
                    "_id": "grant2",
                    "title": "African swallow",
                    "author": "king arthur",
                },
            ],
        ),
    ],
)
def test_merge_collections_superior(input, expected):
    a = input[0]
    b = input[1]
    target_id = "linked_to"
    assert merge_collections_superior(a, b, target_id) == expected


@pytest.mark.parametrize(
    "input,expected",
    [
        (
            (
                [
                    {
                        "_id": "proposal1",
                        "title": "European swallow",
                        "author": "king arthur",
                    }
                ],
                [{"_id": "grant1", "linked_to": "proposal1", "amount": "100 mph"}],
            ),
            [
                {
                    "_id": "grant1",
                    "title": "European swallow",
                    "author": "king arthur",
                    "linked_to": "proposal1",
                    "amount": "100 mph",
                }
            ],
        ),
        (
            (
                [
                    {
                        "_id": "proposal1",
                        "title": "European swallow",
                        "author": "king arthur",
                    },
                    {
                        "_id": "proposal2",
                        "title": "African swallow",
                        "author": "king arthur",
                    },
                ],
                [{"_id": "grant1", "linked_to": "proposal1", "amount": "100 mph"}],
            ),
            [
                {
                    "_id": "grant1",
                    "title": "European swallow",
                    "author": "king arthur",
                    "linked_to": "proposal1",
                    "amount": "100 mph",
                },
            ],
        ),
        (
            (
                [
                    {
                        "_id": "proposal1",
                        "title": "European swallow",
                        "author": "king arthur",
                        "amount": "50 mph",
                    },
                    {
                        "_id": "proposal2",
                        "title": "African swallow",
                        "author": "king arthur",
                    },
                ],
                [{"_id": "grant1", "linked_to": "proposal1", "amount": "100 mph"}],
            ),
            [
                {
                    "_id": "grant1",
                    "title": "European swallow",
                    "author": "king arthur",
                    "linked_to": "proposal1",
                    "amount": "100 mph",
                },
            ],
        ),
        (
            (
                [
                    {
                        "_id": "proposal1",
                        "title": "European swallow",
                        "author": "king arthur",
                        "amount": "50 mph",
                    },
                    {
                        "_id": "proposal2",
                        "title": "African swallow",
                        "author": "king arthur",
                    },
                ],
                [
                    {"_id": "grant1", "linked_to": "proposal1", "amount": "100 mph"},
                ],
            ),
            [
                {
                    "_id": "grant1",
                    "title": "European swallow",
                    "author": "king arthur",
                    "linked_to": "proposal1",
                    "amount": "100 mph",
                },
            ],
        ),
        (
            (
                [
                    {
                        "_id": "proposal1",
                        "title": "European swallow",
                        "author": "king arthur",
                        "amount": "50 mph",
                    },
                    {
                        "_id": "proposal2",
                        "title": "African swallow",
                        "author": "king arthur",
                    },
                ],
                [
                    {"_id": "grant1", "linked_to": "proposal1", "amount": "100 mph"},
                    {
                        "_id": "grant2",
                        "title": "African swallow",
                        "author": "king arthur",
                        "linked_to": "proposal2",
                    },
                ],
            ),
            [
                {
                    "_id": "grant1",
                    "title": "European swallow",
                    "author": "king arthur",
                    "linked_to": "proposal1",
                    "amount": "100 mph",
                },
                {
                    "_id": "grant2",
                    "title": "African swallow",
                    "author": "king arthur",
                    "linked_to": "proposal2",
                },
            ],
        ),
        (
            (
                [
                    {
                        "_id": "proposal1",
                        "title": "European swallow",
                        "author": "king arthur",
                        "amount": "50 mph",
                    },
                ],
                [
                    {"_id": "grant1", "linked_to": "proposal1", "amount": "100 mph"},
                    {
                        "_id": "grant2",
                        "title": "African swallow",
                        "author": "king arthur",
                    },
                ],
            ),
            [
                {
                    "_id": "grant1",
                    "title": "European swallow",
                    "author": "king arthur",
                    "linked_to": "proposal1",
                    "amount": "100 mph",
                },
            ],
        ),
    ],
)
def test_merge_intersection(input, expected):
    a = input[0]
    b = input[1]
    target_id = "linked_to"
    assert merge_collections_intersect(a, b, target_id) == expected


@pytest.mark.parametrize(
    "input,expected,kwargs",
    [
        ("$hi", r"\$hi", {}),
        (
            r"Website: https://github.com/CJ-Wright/" r"Masters_Thesis/raw/master/thesis.pdf hi",
            r"Website: \url{https://github.com/CJ-Wright/" r"Masters_Thesis/raw/master/thesis.pdf} hi",
            {},
        ),
        (
            r"Website: https://github.com/CJ-Wright/" r"Masters_Thesis/raw/master/thesis.pdf hi",
            r"Website: \href{https://github.com/CJ-Wright/" r"Masters_Thesis/raw/master/thesis.pdf} hi",
            {"wrapper": "href"},
        ),
        (
            r"Website: https://github.com/CJ-Wright/" r"Masters_Thesis/raw/master/thesis.pdf hi",
            r"Website: https://github.com/CJ-Wright/" r"Masters\_Thesis/raw/master/thesis.pdf hi",
            {"url_check": False},
        ),
    ],
)
def test_latex_safe(input, expected, kwargs):
    output = latex_safe(input, **kwargs)
    assert output == expected


DEFAULT_SCHEMA = {
    "expenses": {
        "itemized_expenses": {
            "type": "list",
            "schema": {
                "type": "dict",
                "schema": {
                    "day": {
                        "description": "Expense day",
                        "required": True,
                        "type": "integer",
                    },
                },
            },
        },
    },
}

USER_SCHEMA0 = {
    "expenses": {
        "itemized_expenses": {
            "type": "list",
            "schema": {
                "type": "dict",
                "schema": {
                    "day": {
                        "required": False,
                    },
                },
            },
        },
    },
}

EXPECTED_SCHEMA0 = {
    "expenses": {
        "itemized_expenses": {
            "type": "list",
            "schema": {
                "type": "dict",
                "schema": {
                    "day": {
                        "description": "Expense day",
                        "required": False,
                        "type": "integer",
                    },
                },
            },
        },
    },
}

USER_SCHEMA1 = {
    "expenses": {
        "itemized_expenses": {
            "type": "list",
            "schema": {
                "type": "dict",
                "schema": {
                    "day": {
                        "type": "string",
                    },
                },
            },
        },
    },
}

EXPECTED_SCHEMA1 = {
    "expenses": {
        "itemized_expenses": {
            "type": "list",
            "schema": {
                "type": "dict",
                "schema": {
                    "day": {
                        "description": "Expense day",
                        "required": True,
                        "type": "string",
                    },
                },
            },
        },
    },
}

USER_SCHEMA2 = {
    "expenses": {
        "begin_day": {
            "description": "The first day of expense",
            "required": True,
            "type": "string",
        }
    },
}

EXPECTED_SCHEMA2 = {
    "expenses": {
        "itemized_expenses": {
            "type": "list",
            "schema": {
                "type": "dict",
                "schema": {
                    "day": {
                        "description": "Expense day",
                        "required": True,
                        "type": "integer",
                    },
                },
            },
        },
        "begin_day": {
            "description": "The first day of expense",
            "required": True,
            "type": "string",
        },
    },
}

USER_SCHEMA3 = {
    "expenses": {
        "itemized_expenses": {
            "type": "list",
            "schema": {
                "type": "dict",
                "schema": {
                    "day": {"description": "The date on the receipt"},
                },
            },
        },
    },
}

EXPECTED_SCHEMA3 = {
    "expenses": {
        "itemized_expenses": {
            "type": "list",
            "schema": {
                "type": "dict",
                "schema": {
                    "day": {
                        "description": "The date on the receipt",
                        "required": True,
                        "type": "integer",
                    },
                },
            },
        },
    },
}

USER_SCHEMA4 = {
    "expenses": {
        "itemized_expenses": {
            "schema": {
                "schema": {
                    "prepaid_expense": {
                        "description": "Expense paid by the direct billing",
                        "required": True,
                        "type": "float",
                    },
                },
            },
        },
    },
}

EXPECTED_SCHEMA4 = {
    "expenses": {
        "itemized_expenses": {
            "type": "list",
            "schema": {
                "type": "dict",
                "schema": {
                    "day": {
                        "description": "Expense day",
                        "required": True,
                        "type": "integer",
                    },
                    "prepaid_expense": {
                        "description": "Expense paid by the direct billing",
                        "required": True,
                        "type": "float",
                    },
                },
            },
        },
    },
}

USER_SCHEMA5 = {}

EXPECTED_SCHEMA5 = {
    "expenses": {
        "itemized_expenses": {
            "type": "list",
            "schema": {
                "type": "dict",
                "schema": {
                    "day": {
                        "description": "Expense day",
                        "required": True,
                        "type": "integer",
                    },
                },
            },
        },
    },
}

USER_SCHEMA6 = {"expenses": {}}

EXPECTED_SCHEMA6 = {
    "expenses": {
        "itemized_expenses": {
            "type": "list",
            "schema": {
                "type": "dict",
                "schema": {
                    "day": {
                        "description": "Expense day",
                        "required": True,
                        "type": "integer",
                    },
                },
            },
        },
    },
}


@pytest.mark.parametrize(
    "default_schema, user_schema, expected_schema",
    [
        (DEFAULT_SCHEMA, USER_SCHEMA0, EXPECTED_SCHEMA0),
        (DEFAULT_SCHEMA, USER_SCHEMA1, EXPECTED_SCHEMA1),
        (DEFAULT_SCHEMA, USER_SCHEMA2, EXPECTED_SCHEMA2),
        (DEFAULT_SCHEMA, USER_SCHEMA3, EXPECTED_SCHEMA3),
        (DEFAULT_SCHEMA, USER_SCHEMA4, EXPECTED_SCHEMA4),
        (DEFAULT_SCHEMA, USER_SCHEMA5, EXPECTED_SCHEMA5),
        (DEFAULT_SCHEMA, USER_SCHEMA6, EXPECTED_SCHEMA6),
    ],
)
def test_update_schemas(default_schema, user_schema, expected_schema):
    updated_schema = update_schemas(default_schema, user_schema)
    assert updated_schema == expected_schema


def test_group():
    doc0 = {"k0": "v00", "k1": "v01"}
    doc1 = {"k0": "v10", "k1": "v11"}
    doc2 = {"k1": "v21"}
    doc3 = {"k0": "v00", "k1": "v31"}
    db = (doc for doc in (doc0, doc1, doc2, doc3))
    by = "k0"
    expect = {"v00": [doc0, doc3], "v10": [doc1]}
    assert group(db, by) == expect


ppl_coll = [
    {
        "_id": "m1",
        "name": "member1",
        "education": [
            {"group": "bg", "institution": "columbiau", "degree": "PhD", "department": "apam", "begin_year": 2016}
        ],
        "employment": [
            {
                "begin_year": 2020,
                "begin_month": 1,
                "organization": "columbiau",
                "position": "Undergraduate Researcher",
                "advisor": "sbillinge",
                "status": "undergrad",
            }
        ],
    },
    {
        "_id": "nm1",
        "name": "non-member1",
        "education": [{"institution": "columbiau", "degree": "PhD", "department": "apam", "begin_year": 2016}],
        "employment": [
            {
                "begin_year": 2020,
                "begin_month": 1,
                "organization": "columbiau",
                "position": "Undergraduate Researcher",
                "advisor": "sbillinge",
                "status": "undergrad",
            }
        ],
    },
    {
        "_id": "m2",
        "name": "member2",
        "education": [{"institution": "columbiau", "degree": "PhD", "department": "apam", "begin_year": 2016}],
        "employment": [
            {
                "begin_year": 2020,
                "begin_month": 1,
                "group": "bg",
                "organization": "columbiau",
                "position": "Undergraduate Researcher",
                "advisor": "sbillinge",
                "status": "undergrad",
            }
        ],
    },
]


@pytest.mark.parametrize(
    "input,expected",
    [
        (ppl_coll, set(["m1", "m2"])),
    ],
)
def test_group_member_ids(input, expected):
    actual = group_member_ids(input, "bg")
    assert actual == expected


d1 = {
    "name": "John",
    "experience": [{"company": "Google", "role": "product manager"}, {"company": "Amazon", "role": "QA"}],
    "school": {"name": "Columbia", "location": "NYC", "year": "senior"},
}
d2 = {
    "name": "Sarah",
    "experience": [{"company": "Verizon", "role": "sales"}, {"company": "AT&T", "role": "software engineer"}],
    "school": {"name": "Columbia", "location": "NYC", "year": "junior"},
    "hobbies": ["swimming", "hiking"],
    "info": {"stats": {"code": "a76", "location": "California"}, "software": "CAD"},
}
d3 = {
    "_id": "abc",
    "name": "Example Lab",
    "Members": [
        {
            "Name": "Lisa",
            "Experience": [
                {"company": "Google", "location": {"state": "CA", "zip code": "94043"}},
                {"company": "Amazon", "location": {"state": "VA", "zip code": "20189"}},
            ],
        },
        {
            "Name": "Stephen",
            "Experience": [{"company": "Goldman Sachs", "location": {"state": "NY", "zip code": "10282"}}],
        },
    ],
}


@pytest.mark.parametrize(
    "input,expected",
    [
        (d1, ["John", "Google", "product manager", "Amazon", "QA", "Columbia", "NYC", "senior"]),
        (
            d2,
            [
                "Sarah",
                "Verizon",
                "sales",
                "AT&T",
                "software engineer",
                "Columbia",
                "NYC",
                "junior",
                "swimming",
                "hiking",
                "a76",
                "California",
                "CAD",
            ],
        ),
        (
            d3,
            [
                "abc",
                "Example Lab",
                "Lisa",
                "Google",
                "CA",
                "94043",
                "Amazon",
                "VA",
                "20189",
                "Stephen",
                "Goldman Sachs",
                "NY",
                "10282",
            ],
        ),
    ],
)
def test_compound_dict(input, expected):
    assert compound_dict(input, []) == expected


l1 = ["hello", {"name": "Fred", "status": "active"}, {"name": "Derf", "status": "inactive"}, "bye"]
l2 = [
    ["a", "b", "c"],
    {"name": "Anthony", "Status": "active"},
    [{"product": "phone"}, {"product": "laptop"}],
    "end",
]


@pytest.mark.parametrize(
    "input,expected",
    [
        (l1, ["hello", "Fred", "active", "Derf", "inactive", "bye"]),
        (l2, ["a", "b", "c", "Anthony", "active", "phone", "laptop", "end"]),
    ],
)
def test_compound_list(input, expected):
    assert compound_list(input, []) == expected


p1 = {
    "_id": "scopatz",
    "aka": [
        "Scopatz",
        "Scopatz, A",
        "Scopatz, A.",
        "Scopatz, A M",
        "Anthony Michael Scopatz",
    ],
    "name": "Anthony Scopatz",
}
p2 = {
    "_id": "abc",
    "aka": [
        "A. BC",
        "BC, A",
        "Anthony BC",
    ],
    "name": "Anthony Bill Chris",
}
p3 = {
    "_id": "Leonan",
    "name": "Leonan Garea",
    "company": {"name": "Amazon", "role": "Product Manager"},
    "projects": [
        {"title": "GUI Application", "description": "Make a GUI for the group"},
    ],
}
p4 = {
    "_id": "cba",
    "name": "Jackie",
    "company": {},
    "projects": [
        {"title": "PDF Maker", "description": "New PDF function"},
        {"title": "Write Paper", "description": "Draft the new paper"},
    ],
}
p5 = {
    "_id": "ghi",
    "name": "Carl",
    "experience": [
        {
            "name": "Google",
            "roles": [
                {"position": "software engineer", "location": {"state": "CA", "zip code": "92551"}},
                {"position": "manager", "location": {"state": "VA", "zip code": "20189"}},
            ],
        },
        {
            "name": "Goldman Sachs",
            "Experience": [{"position": "junior associate", "location": {"state": "NY", "zip code": "10282"}}],
        },
    ],
}


@pytest.mark.parametrize(
    "input, expected",
    [
        (([p1, p2], ["aka", "name", "_id"], "Anth", False), [p1, p2]),
        (([p1, p2], ["aka", "name", "_id"], "scopatz, a", True), []),
        (([p1, p2], ["aka", "name", "_id"], "scopatz, a", False), [p1]),
        (([p1, p2], ["aka", "name", "_id"], "ill", False), [p2]),
        (([p3], ["company"], "Amazon", False), [p3]),
        (([p3, p4], ["projects"], "PDF", False), [p4]),
        (([p5], ["experience"], "20189", False), [p5]),
        (([p5], ["experience"], "hello", False), []),
    ],
)
def test_fragment_retrieval(input, expected):
    assert fragment_retrieval(input[0], input[1], input[2], case_sensitive=input[3]) == expected


@pytest.mark.parametrize(
    "input, expected",
    [
        ((None, None), "present"),
        ((None, 2002), "2002"),
        ((5, 2002), "May 2002"),
    ],
)
def test_month_and_year(input, expected):
    assert month_and_year(input[0], input[1]) == expected


@pytest.mark.parametrize(
    "appts,start,end,expected",
    [
        (
            {
                "name": "Kurt Godel",
                "_id": "kgodel",
                "appointments": {
                    "A": {
                        "begin_year": 2017,
                        "begin_month": 6,
                        "begin_day": 1,
                        "end_year": 2017,
                        "end_month": 6,
                        "end_day": 30,
                        "grant": "grant1",
                        "loading": 1.0,
                        "type": "pd",
                    }
                },
            },
            "2017-06-01",
            "2017-07-01",
            False,
        ),
        (
            {
                "name": "MC Escher",
                "_id": "mcescher",
                "appointments": {
                    "A": {
                        "begin_date": "2017-06-01",
                        "end_date": "2017-06-30",
                        "grant": "grant1",
                        "loading": 0.5,
                        "type": "pd",
                    },
                    "B": {
                        "begin_date": "2017-06-01",
                        "end_date": "2017-06-30",
                        "grant": "grant2",
                        "loading": 0.5,
                        "type": "pd",
                    },
                },
            },
            "2017-06-01",
            "2017-06-30",
            True,
        ),
        (
            {
                "name": "Johann Sebastian Bach",
                "_id": "jsbach",
                "appointments": {
                    "A": {
                        "begin_date": "2017-06-01",
                        "end_date": "2017-06-30",
                        "grant": "grant1",
                        "loading": 0.5,
                        "type": "pd",
                    },
                    "B": {
                        "begin_date": "2017-06-02",
                        "end_date": "2017-06-29",
                        "grant": "grant2",
                        "loading": 0.5,
                        "type": "pd",
                    },
                },
            },
            "2017-06-01",
            "2017-06-30",
            False,
        ),
        (
            {
                "name": "Evariste Galois",
                "_id": "egalois",
                "appointments": {
                    "A": {
                        "begin_date": "2017-06-01",
                        "end_date": "2017-06-15",
                        "grant": "grant1",
                        "loading": 1.0,
                        "type": "pd",
                    },
                    "B": {
                        "begin_date": "2017-06-16",
                        "end_date": "2017-06-30",
                        "grant": "grant2",
                        "loading": 1.0,
                        "type": "pd",
                    },
                },
            },
            "2017-06-01",
            "2017-06-30",
            True,
        ),
        (
            {
                "name": "Ludwig Wittgenstein",
                "_id": "lwittgenstein",
                "appointments": {
                    "A": {
                        "begin_date": "2017-06-01",
                        "end_date": "2017-06-15",
                        "grant": "grant1",
                        "loading": 1.0,
                        "type": "pd",
                    },
                    "B": {
                        "begin_date": "2017-06-17",
                        "end_date": "2017-06-30",
                        "grant": "grant2",
                        "loading": 1.0,
                        "type": "pd",
                    },
                    "C": {
                        "begin_date": "2017-07-01",
                        "end_date": "2017-07-30",
                        "grant": "grant3",
                        "loading": 1.0,
                        "type": "pd",
                    },
                },
            },
            "2017-06-01",
            "2017-06-30",
            False,
        ),
        (
            {
                "name": "Buckminster Fuller",
                "_id": "bfuller",
                "appointments": {
                    "A": {
                        "begin_date": "2017-06-01",
                        "end_date": "2017-06-30",
                        "grant": "grant1",
                        "loading": 1.0,
                        "type": "pd",
                    },
                    "B": {
                        "begin_date": "2017-06-17",
                        "end_date": "2017-06-30",
                        "grant": "grant2",
                        "loading": 1.0,
                        "type": "pd",
                    },
                },
            },
            "2017-06-01",
            "2017-06-30",
            False,
        ),
        (
            {
                "name": "Lorem Ipsum",
                "_id": "lipsum",
                "appointments": {
                    "A": {
                        "begin_date": "2017-06-01",
                        "end_date": "2017-06-30",
                        "grant": "grant1",
                        "loading": 1.0,
                        "type": "pd",
                    },
                    "B": {
                        "begin_date": "2017-06-17",
                        "end_date": "2017-06-30",
                        "grant": "grant2",
                        "loading": 1.0,
                        "type": "pd",
                    },
                },
            },
            "2017-06-01",
            "2017-06-30",
            False,
        ),
    ],
)
def test_is_fully_appointed(appts, start, end, expected):
    actual = is_fully_appointed(appts, start, end)
    assert actual == expected


@pytest.mark.parametrize(
    "input, expected",
    [
        (
            "honors",
            [
                {"description": "Omega Laser User's Group Travel Award (\\$1,100)", "year": 2013, "_key": 2013.0},
            ],
        ),
        (
            "service",
            [
                {"_key": 2020.01, "description": "International Steering Committee", "year": 2020},
                {"description": "Omega Laser User's Group Travel Award (\\$1,100)", "year": 2013, "_key": 2013.0},
            ],
        ),
    ],
)
def test_awards_grants_honors(input, expected):
    assert awards_grants_honors(PEOPLE_COLL[0], input, funding=True, service_types=None) == expected


@pytest.mark.parametrize(
    "input, expected",
    [
        (([{"_id": "afriend", "aka": ["AB Friend", "Tony Friend"], "name": "Anthony B Friend"}], "Simon"), None),
        (
            (
                [{"_id": "afriend", "aka": ["AB Friend", "Tony Friend"], "name": "Anthony B Friend"}],
                "Anthony B Friend",
            ),
            "afriend",
        ),
        (
            (
                [
                    {"_id": "afriend", "aka": ["AB Friend", "Tony Friend"], "name": "Anthony B Friend"},
                    {"_id": "aeinstein", "aka": ["Einstein"], "name": "Albert Einstein"},
                ],
                "Albert Einstein",
            ),
            "aeinstein",
        ),
    ],
)
def test_get_id_from_name(input, expected):
    assert get_id_from_name(input[0], input[1]) == expected


@pytest.mark.parametrize(
    "input, expected",
    [
        ((2012, "Jan", 18), "Wed, 18 Jan 2012 00:00:00 -0000"),
        ((2020, 6, 22), "Mon, 22 Jun 2020 00:00:00 -0000"),
    ],
)
def test_date_to_rfc822(input, expected):
    assert date_to_rfc822(input[0], input[1], input[2]) == expected


person1 = {
    "_id": "scopatz",
    "aka": [
        "Scopatz",
        "Scopatz, A",
        "Scopatz, A.",
        "Scopatz, A M",
        "Anthony Michael Scopatz",
    ],
    "name": "Anthony Scopatz",
    "position": "Professor",
}
person2 = {
    "_id": "abc",
    "aka": [
        "A. BC",
        "BC, A",
        "Anthony BC",
    ],
    "name": "Anthony Bill Chris",
    "position": "Professor",
}
person3 = {
    "_id": "jdoe",
    "aka": [
        "A. BC",
        "BC, A",
        "Anthony BC",
    ],
    "name": "John Doe",
}
people = [person1, person2, person3]


@pytest.mark.parametrize(
    "input, expected",
    [
        ((people, ["name", "Doe"]), [person3]),
        ((people, ["name", "Jerry"]), []),
        ((people, ["position", "Prof"]), [person1, person2]),
        ((people, ["position", "Prof", "name", "Chris"]), [person2]),
    ],
)
def test_key_value_pair_filter(input, expected):
    assert key_value_pair_filter(input[0], input[1]) == expected


@pytest.mark.parametrize(
    "input, expected",
    [
        (([person3], None), "jdoe    \n"),
        (([], None), ""),
        (
            ([person1, person2], ["position"]),
            "scopatz    position: Professor    \nabc    position: Professor    \n",
        ),
        (([person2], ["position"]), "abc    position: Professor    \n"),
    ],
)
def test_collection_str(input, expected):
    assert collection_str(input[0], input[1]) == expected


@pytest.mark.parametrize(
    "input, expected",
    [
        ((people, ["name", "Doe"], None), "jdoe    \n"),
        ((people, ["name", "Jerry"], None), ""),
        ((people, ["position", "Prof", "name", "Chris"], None), "abc    \n"),
        ((people, ["position", "prof", "name", "Chris"], ["position"]), "abc    position: Professor    \n"),
    ],
)
def test_search_collection(input, expected):
    assert search_collection(input[0], input[1], input[2]) == expected


appointed_people = [
    {
        "name": "Kurt Godel",
        "_id": "kgodel",
        "appointments": {
            "A": {
                "begin_date": "2019-09-01",
                "end_date": "2019-09-10",
                "grant": "grant1",
                "loading": 0.5,
                "type": "gra",
            },
            "B": {
                "_id": "B",
                "begin_date": "2019-09-01",
                "end_date": "2019-09-10",
                "grant": "grant1",
                "loading": 0.25,
                "type": "gra",
            },
            "C": {
                "_id": "C",
                "begin_date": "2019-09-01",
                "end_date": "2019-09-10",
                "grant": "grant1",
                "loading": 0.25,
                "type": "gra",
            },
        },
        "employment": [
            {"group": "permutation", "begin_date": "2014-06-01", "end_date": "2015-06-01", "status": "phd"},
            {"group": "matrix", "begin_year": "2020", "end_day": "5", "end_month": "12", "end_year": "2020"},
            {
                "group": "permutation",
                "begin_day": 4,
                "begin_month": 9,
                "begin_year": 2012,
                "end_day": 5,
                "end_month": 9,
                "end_year": 2012,
                "permanent": "true",
            },
        ],
    },
    {
        "name": "MC Escher",
        "_id": "mcescher",
        "appointments": {
            "A": {
                "begin_date": "2019-10-01",
                "end_date": "2019-10-31",
                "grant": "grant1",
                "loading": 1.0,
                "type": "ss",
            },
            "B": {
                "begin_date": "2019-11-01",
                "end_date": "2019-11-30",
                "grant": "grant2",
                "loading": 0.5,
                "type": "ss",
            },
            "C": {
                "begin_date": "2019-11-01",
                "end_date": "2019-11-30",
                "grant": "grant3",
                "loading": 0.5,
                "type": "ss",
            },
        },
        "employment": [
            {
                "group": "transformation",
                "begin_date": "2018-07-24",
                "end_date": dt.date(2020, 8, 1),
                "status": "postdoc",
            },
            {"group": "abstract", "begin_year": 2010, "end_day": 5, "end_month": 12, "end_year": 2020},
            {"group": "abstract", "begin_date": "2012-06-30", "end_date": "2012-09-05"},
        ],
    },
    {
        "name": "Johann Sebastian Bach",
        "_id": "jsbach",
        "appointments": {
            "A": {
                "begin_date": "2019-12-01",
                "end_date": "2020-12-15",
                "grant": "grant1",
                "loading": 0.9,
                "type": "pd",
            },
            "B": {
                "begin_date": "2019-12-16",
                "end_date": "2020-12-31",
                "grant": "grant2",
                "loading": 0.9,
                "type": "pd",
            },
            "C": {
                "begin_date": "2019-12-01",
                "end_date": "2020-12-31",
                "grant": "grant3",
                "loading": 0.1,
                "type": "pd",
            },
        },
        "employment": [{"group": "bg", "begin_date": "2019-02-03"}],
    },
    {
        "name": "Ludwig Wittgenstein",
        "_id": "lwittgenstein",
        "appointments": {
            "A": {
                "begin_date": "2019-12-10",
                "end_date": "2019-12-20",
                "grant": "grant2",
                "loading": 1.0,
                "type": "ss",
            }
        },
    },
    {
        "name": "Karl Popper",
        "_id": "kpopper",
        "appointments": {
            "A": {
                "begin_date": "2019-12-25",
                "end_date": "2019-12-31",
                "grant": "grant2",
                "loading": 1.0,
                "type": "ss",
            }
        },
    },
    {"name": "GEM Anscombe", "_id": "ganscombe", "appointments": {}},
    {
        "name": "Sophie Germain",
        "_id": "sgermain",
        "appointments": {
            "A": {
                "begin_date": "2019-09-02",
                "end_date": "2019-09-06",
                "grant": "grant4",
                "loading": 1.0,
                "type": "ss",
            }
        },
    },
]


@pytest.mark.parametrize(
    "people,key,value,start,end,expected",
    [
        (
            appointed_people,
            "grant",
            "grant1",
            None,
            None,
            [
                {
                    "person": "kgodel",
                    "_id": "A",
                    "begin_date": "2019-09-01",
                    "end_date": "2019-09-10",
                    "grant": "grant1",
                    "loading": 0.5,
                    "type": "gra",
                },
                {
                    "person": "kgodel",
                    "_id": "B",
                    "begin_date": "2019-09-01",
                    "end_date": "2019-09-10",
                    "grant": "grant1",
                    "loading": 0.25,
                    "type": "gra",
                },
                {
                    "person": "kgodel",
                    "_id": "C",
                    "begin_date": "2019-09-01",
                    "end_date": "2019-09-10",
                    "grant": "grant1",
                    "loading": 0.25,
                    "type": "gra",
                },
                {
                    "person": "mcescher",
                    "_id": "A",
                    "begin_date": "2019-10-01",
                    "end_date": "2019-10-31",
                    "grant": "grant1",
                    "loading": 1.0,
                    "type": "ss",
                },
                {
                    "person": "jsbach",
                    "_id": "A",
                    "begin_date": "2019-12-01",
                    "end_date": "2020-12-15",
                    "grant": "grant1",
                    "loading": 0.9,
                    "type": "pd",
                },
            ],
        ),
        (
            appointed_people,
            None,
            None,
            "2019-09-01",
            "2019-09-30",
            [
                {
                    "person": "kgodel",
                    "_id": "A",
                    "begin_date": "2019-09-01",
                    "end_date": "2019-09-10",
                    "grant": "grant1",
                    "loading": 0.5,
                    "type": "gra",
                },
                {
                    "person": "kgodel",
                    "_id": "B",
                    "begin_date": "2019-09-01",
                    "end_date": "2019-09-10",
                    "grant": "grant1",
                    "loading": 0.25,
                    "type": "gra",
                },
                {
                    "person": "kgodel",
                    "_id": "C",
                    "begin_date": "2019-09-01",
                    "end_date": "2019-09-10",
                    "grant": "grant1",
                    "loading": 0.25,
                    "type": "gra",
                },
                {
                    "person": "sgermain",
                    "_id": "A",
                    "begin_date": "2019-09-02",
                    "end_date": "2019-09-06",
                    "grant": "grant4",
                    "loading": 1.0,
                    "type": "ss",
                },
            ],
        ),
        (
            appointed_people,
            ["loading", "type"],
            [1.0, "ss"],
            "2019-12-15",
            "2019-12-25",
            [
                {
                    "person": "lwittgenstein",
                    "_id": "A",
                    "begin_date": "2019-12-10",
                    "end_date": "2019-12-20",
                    "grant": "grant2",
                    "loading": 1.0,
                    "type": "ss",
                },
                {
                    "person": "kpopper",
                    "_id": "A",
                    "begin_date": "2019-12-25",
                    "end_date": "2019-12-31",
                    "grant": "grant2",
                    "loading": 1.0,
                    "type": "ss",
                },
            ],
        ),
        (appointed_people, ["loading", "type", "grant"], [0.9, "pd", "grant3"], None, None, []),
        (
            appointed_people,
            None,
            None,
            None,
            None,
            [
                {
                    "person": "kgodel",
                    "_id": "A",
                    "begin_date": "2019-09-01",
                    "end_date": "2019-09-10",
                    "grant": "grant1",
                    "loading": 0.5,
                    "type": "gra",
                },
                {
                    "person": "kgodel",
                    "_id": "B",
                    "begin_date": "2019-09-01",
                    "end_date": "2019-09-10",
                    "grant": "grant1",
                    "loading": 0.25,
                    "type": "gra",
                },
                {
                    "person": "kgodel",
                    "_id": "C",
                    "begin_date": "2019-09-01",
                    "end_date": "2019-09-10",
                    "grant": "grant1",
                    "loading": 0.25,
                    "type": "gra",
                },
                {
                    "person": "mcescher",
                    "_id": "A",
                    "begin_date": "2019-10-01",
                    "end_date": "2019-10-31",
                    "grant": "grant1",
                    "loading": 1.0,
                    "type": "ss",
                },
                {
                    "person": "mcescher",
                    "_id": "B",
                    "begin_date": "2019-11-01",
                    "end_date": "2019-11-30",
                    "grant": "grant2",
                    "loading": 0.5,
                    "type": "ss",
                },
                {
                    "person": "mcescher",
                    "_id": "C",
                    "begin_date": "2019-11-01",
                    "end_date": "2019-11-30",
                    "grant": "grant3",
                    "loading": 0.5,
                    "type": "ss",
                },
                {
                    "person": "jsbach",
                    "_id": "A",
                    "begin_date": "2019-12-01",
                    "end_date": "2020-12-15",
                    "grant": "grant1",
                    "loading": 0.9,
                    "type": "pd",
                },
                {
                    "person": "jsbach",
                    "_id": "B",
                    "begin_date": "2019-12-16",
                    "end_date": "2020-12-31",
                    "grant": "grant2",
                    "loading": 0.9,
                    "type": "pd",
                },
                {
                    "person": "jsbach",
                    "_id": "C",
                    "begin_date": "2019-12-01",
                    "end_date": "2020-12-31",
                    "grant": "grant3",
                    "loading": 0.1,
                    "type": "pd",
                },
                {
                    "person": "lwittgenstein",
                    "_id": "A",
                    "begin_date": "2019-12-10",
                    "end_date": "2019-12-20",
                    "grant": "grant2",
                    "loading": 1.0,
                    "type": "ss",
                },
                {
                    "person": "kpopper",
                    "_id": "A",
                    "begin_date": "2019-12-25",
                    "end_date": "2019-12-31",
                    "grant": "grant2",
                    "loading": 1.0,
                    "type": "ss",
                },
                {
                    "person": "sgermain",
                    "_id": "A",
                    "begin_date": "2019-09-02",
                    "end_date": "2019-09-06",
                    "grant": "grant4",
                    "loading": 1.0,
                    "type": "ss",
                },
            ],
        ),
        (appointed_people, "type", "ss", "2019-10-21", "2019-09-01", "begin date is after end date"),
        (
            appointed_people,
            ["type", "loading"],
            None,
            None,
            None,
            "number of filter keys and filter values do not match",
        ),
        (
            appointed_people,
            "type",
            "pd",
            "2019-12-10",
            None,
            "please enter both begin date and end date or neither",
        ),
        (
            [
                {
                    "name": "Magical Person",
                    "_id": "mperson",
                    "appointments": {
                        "A": {
                            "begin_date": "2019-09-01",
                            "end_date": "2019-09-05",
                            "loading": 1.0,
                            "grant": "grant1",
                            "type": "imaginary",
                        }
                    },
                }
            ],
            None,
            None,
            None,
            None,
            "invalid  type imaginary for appointment A of mperson",
        ),
    ],
)
def test_collect_appts(people, key, value, start, end, expected):
    try:
        actual = collect_appts(people, filter_key=key, filter_value=value, begin_date=start, end_date=end)
        assert actual == expected
    except ValueError:
        with pytest.raises(ValueError) as excinfo:
            actual = collect_appts(people, filter_key=key, filter_value=value, begin_date=start, end_date=end)
        assert str(excinfo.value) == expected
    except RuntimeError:
        with pytest.raises(RuntimeError) as excinfo:
            actual = collect_appts(people, filter_key=key, filter_value=value, begin_date=start, end_date=end)
        assert str(excinfo.value) == expected


appts = collect_appts(appointed_people)
grant1 = {
    "_id": "grant1",
    "alias": "grant_one",
    "budget": [
        {
            "begin_date": "2019-09-01",
            "end_date": "2019-09-03",
            "student_months": 1,
            "postdoc_months": 0.5,
            "ss_months": 0,
        },
        {
            "begin_date": "2019-09-04",
            "end_date": "2019-09-07",
            "student_months": 1.5,
            "postdoc_months": 0,
            "ss_months": 0,
        },
        {
            "begin_date": "2019-09-08",
            "end_date": "2019-09-10",
            "student_months": 2,
            "postdoc_months": 1.5,
            "ss_months": 0,
        },
    ],
}
grant2 = {
    "_id": "grant2",
    "alias": "grant_two",
    "budget": [
        {
            "begin_date": "2019-09-01",
            "end_date": "2019-12-31",
            "student_months": 4,
            "postdoc_months": 2.5,
            "ss_months": 1,
        }
    ],
}
grant3 = {
    "_id": "grant3",
    "budget": [
        {
            "begin_date": "2019-09-01",
            "end_date": "2019-10-31",
            "student_months": 0,
            "postdoc_months": 1,
            "ss_months": 2,
        },
        {
            "begin_date": "2019-11-01",
            "end_date": "2019-12-31",
            "student_months": 2,
            "postdoc_months": 0.5,
            "ss_months": 0,
        },
    ],
}
grant4 = {
    "_id": "grant4",
    "alias": "grant_four",
    "budget": [
        {
            "begin_date": "2019-09-01",
            "end_date": "2019-09-07",
            "student_months": 1,
            "postdoc_months": 1,
            "ss_months": 1,
        }
    ],
}


@pytest.mark.parametrize(
    "grant,appointments,start,end,expected",
    [
        (
            grant1,
            appts,
            None,
            None,
            {
                dt.date(2019, 9, 1): {"postdoc_days": 15.25, "ss_days": 0.0, "student_days": 29.5},
                dt.date(2019, 9, 2): {"postdoc_days": 15.25, "ss_days": 0.0, "student_days": 28.5},
                dt.date(2019, 9, 3): {"postdoc_days": 15.25, "ss_days": 0.0, "student_days": 27.5},
                dt.date(2019, 9, 4): {"postdoc_days": 15.25, "ss_days": 0.0, "student_days": 72.25},
                dt.date(2019, 9, 5): {"postdoc_days": 15.25, "ss_days": 0.0, "student_days": 71.25},
                dt.date(2019, 9, 6): {"postdoc_days": 15.25, "ss_days": 0.0, "student_days": 70.25},
                dt.date(2019, 9, 7): {"postdoc_days": 15.25, "ss_days": 0.0, "student_days": 69.25},
                dt.date(2019, 9, 8): {"postdoc_days": 61.0, "ss_days": 0.0, "student_days": 129.25},
                dt.date(2019, 9, 9): {"postdoc_days": 61.0, "ss_days": 0.0, "student_days": 128.25},
                dt.date(2019, 9, 10): {"postdoc_days": 61.0, "ss_days": 0.0, "student_days": 127.25},
            },
        ),
        (
            grant2,
            appts,
            "2019-12-15",
            "2019-12-31",
            {
                dt.date(2019, 12, 15): {"postdoc_days": 76.25, "ss_days": 9.5, "student_days": 122.0},
                dt.date(2019, 12, 16): {"postdoc_days": 75.35, "ss_days": 8.5, "student_days": 122.0},
                dt.date(2019, 12, 17): {"postdoc_days": 74.45, "ss_days": 7.5, "student_days": 122.0},
                dt.date(2019, 12, 18): {"postdoc_days": 73.55, "ss_days": 6.5, "student_days": 122.0},
                dt.date(2019, 12, 19): {"postdoc_days": 72.65, "ss_days": 5.5, "student_days": 122.0},
                dt.date(2019, 12, 20): {"postdoc_days": 71.75, "ss_days": 4.5, "student_days": 122.0},
                dt.date(2019, 12, 21): {"postdoc_days": 70.85, "ss_days": 4.5, "student_days": 122.0},
                dt.date(2019, 12, 22): {"postdoc_days": 69.95, "ss_days": 4.5, "student_days": 122.0},
                dt.date(2019, 12, 23): {"postdoc_days": 69.05, "ss_days": 4.5, "student_days": 122.0},
                dt.date(2019, 12, 24): {"postdoc_days": 68.15, "ss_days": 4.5, "student_days": 122.0},
                dt.date(2019, 12, 25): {"postdoc_days": 67.25, "ss_days": 3.5, "student_days": 122.0},
                dt.date(2019, 12, 26): {"postdoc_days": 66.35, "ss_days": 2.5, "student_days": 122.0},
                dt.date(2019, 12, 27): {"postdoc_days": 65.45, "ss_days": 1.5, "student_days": 122.0},
                dt.date(2019, 12, 28): {"postdoc_days": 64.55, "ss_days": 0.5, "student_days": 122.0},
                dt.date(2019, 12, 29): {"postdoc_days": 63.65, "ss_days": -0.5, "student_days": 122.0},
                dt.date(2019, 12, 30): {"postdoc_days": 62.75, "ss_days": -1.5, "student_days": 122.0},
                dt.date(2019, 12, 31): {"postdoc_days": 61.85, "ss_days": -2.5, "student_days": 122.0},
            },
        ),
        (
            grant3,
            appts,
            "2019-12-31",
            "2019-12-31",
            {dt.date(2019, 12, 31): {"postdoc_days": 42.65, "ss_days": 46.0, "student_days": 61.0}},
        ),
        (
            grant4,
            appts,
            None,
            None,
            {
                dt.date(2019, 9, 1): {"postdoc_days": 30.5, "ss_days": 30.5, "student_days": 30.5},
                dt.date(2019, 9, 2): {"postdoc_days": 30.5, "ss_days": 29.5, "student_days": 30.5},
                dt.date(2019, 9, 3): {"postdoc_days": 30.5, "ss_days": 28.5, "student_days": 30.5},
                dt.date(2019, 9, 4): {"postdoc_days": 30.5, "ss_days": 27.5, "student_days": 30.5},
                dt.date(2019, 9, 5): {"postdoc_days": 30.5, "ss_days": 26.5, "student_days": 30.5},
                dt.date(2019, 9, 6): {"postdoc_days": 30.5, "ss_days": 25.5, "student_days": 30.5},
                dt.date(2019, 9, 7): {"postdoc_days": 30.5, "ss_days": 25.5, "student_days": 30.5},
            },
        ),
        (
            {"_id": "magical_grant", "alias": "very_magical_grant"},
            appts,
            "2012-12-23",
            "2013-01-24",
            "magical_grant has no specified budget",
        ),
        (
            grant4,
            appointed_people[0].get("appointments"),
            None,
            None,
            {
                dt.date(2019, 9, 1): {"postdoc_days": 30.5, "ss_days": 30.5, "student_days": 30.5},
                dt.date(2019, 9, 2): {"postdoc_days": 30.5, "ss_days": 30.5, "student_days": 30.5},
                dt.date(2019, 9, 3): {"postdoc_days": 30.5, "ss_days": 30.5, "student_days": 30.5},
                dt.date(2019, 9, 4): {"postdoc_days": 30.5, "ss_days": 30.5, "student_days": 30.5},
                dt.date(2019, 9, 5): {"postdoc_days": 30.5, "ss_days": 30.5, "student_days": 30.5},
                dt.date(2019, 9, 6): {"postdoc_days": 30.5, "ss_days": 30.5, "student_days": 30.5},
                dt.date(2019, 9, 7): {"postdoc_days": 30.5, "ss_days": 30.5, "student_days": 30.5},
            },
        ),
    ],
)
def test_grant_burn(grant, appointments, start, end, expected):
    try:
        actual = grant_burn(grant, appointments, begin_date=start, end_date=end)
        assert actual == expected
    except ValueError:
        with pytest.raises(ValueError) as excinfo:
            actual = grant_burn(grant, appointments, begin_date=start, end_date=end)
        assert str(excinfo.value) == expected


meeting1 = {"_id": "grp2020-06-15", "journal_club": {"doi": "TBD"}}
meeting2 = {"_id": "grp2020-06-22", "presentation": {"link": "TBD"}}
meeting3 = {"_id": "grp2020-06-29", "presentation": {"link": "2002ak_grmtg_presnetation", "title": "tbd"}}


@pytest.mark.parametrize(
    "meeting,date,expected",
    [
        (meeting1, dt.date(2020, 8, 15), "grp2020-06-15 does not have a journal club doi"),
        (meeting1, dt.date(2020, 5, 15), None),
        (meeting2, dt.date(2020, 8, 15), "grp2020-06-22 does not have a presentation link"),
        (meeting2, dt.date(2020, 5, 15), None),
        (meeting3, dt.date(2020, 8, 15), "grp2020-06-29 does not have a presentation title"),
        (meeting3, dt.date(2020, 5, 15), None),
    ],
)
def test_validate_meeting(meeting, date, expected):
    try:
        actual = validate_meeting(meeting, date)
        assert actual == expected
    except ValueError:
        with pytest.raises(ValueError) as excinfo:
            actual = validate_meeting(meeting, date)
        assert str(excinfo.value) == expected


@pytest.mark.parametrize(
    "person,grpname,expected",
    [
        (
            appointed_people[0],
            "permutation",
            [
                {
                    "_id": "kgodel",
                    "begin_date": dt.date(2014, 6, 1),
                    "end_date": dt.date(2015, 6, 1),
                    "status": "phd",
                    "permanent": None,
                },
                {
                    "_id": "kgodel",
                    "begin_date": dt.date(2012, 9, 4),
                    "end_date": dt.date(2012, 9, 5),
                    "status": None,
                    "permanent": "true",
                },
            ],
        ),
        (
            appointed_people[1],
            "transformation",
            [
                {
                    "_id": "mcescher",
                    "begin_date": dt.date(2018, 7, 24),
                    "end_date": dt.date(2020, 8, 1),
                    "status": "postdoc",
                    "permanent": None,
                }
            ],
        ),
        (appointed_people[2], "bg", "WARNING: jsbach has no end date in employment for bg starting 2019-02-03"),
        (appointed_people[3], "abstract", []),
    ],
)
def test_group_member_employment_start_end(person, grpname, expected):
    try:
        actual = group_member_employment_start_end(person, grpname)
        assert actual == expected
    except RuntimeError:
        with pytest.raises(RuntimeError) as excinfo:
            actual = group_member_employment_start_end(person, grpname)
        assert str(excinfo.value) == expected


@pytest.mark.parametrize(
    "inp,expected",
    [
        ([{"dupe_key": 1, "nond": 2}], [{"dupe_key": 1, "nond": 2}]),
        ([{"dupe_key": 1, "nond": 2}, {"dupe_key": 1, "nond": 3}], [{"dupe_key": 1, "nond": 2}]),
        (
            [{"no_dupe_key": 1, "nond": 2}],
            "ERROR: Target key, dupe_key not found in {'no_dupe_key': 1, 'nond': 2}",
        ),
    ],
)
def test_remove_duplicate_docs(inp, expected):
    try:
        actual = remove_duplicate_docs(inp, "dupe_key")
        assert actual == expected
    except RuntimeError:
        with pytest.raises(RuntimeError) as excinfo:
            actual = remove_duplicate_docs(inp, "dupe_key")
        assert str(excinfo.value) == expected


@pytest.mark.parametrize(
    "inp,expected",
    [
        (
            [
                [
                    {
                        "_id": "student",
                        "name": "Lancelot",
                        "employment": [
                            {
                                "status": "ms",
                                "begin_date": "2020-05-05",
                                "end_date": "2020-10-10",
                                "advisor": "awesome",
                                "position": "masters researcher",
                            }
                        ],
                    }
                ],
                "2022-01-01",
            ],
            [],
        ),
        (
            [
                [
                    {
                        "_id": "student",
                        "name": "Lancelot",
                        "employment": [
                            {
                                "status": "ms",
                                "begin_date": "2020-05-05",
                                "end_date": "2020-10-10",
                                "advisor": "awesome",
                                "position": "masters researcher",
                            }
                        ],
                    }
                ],
                "2019-01-01",
            ],
            [
                {
                    "_id": "student",
                    "name": "Lancelot",
                    "employment": [
                        {
                            "status": "ms",
                            "begin_date": "2020-05-05",
                            "end_date": "2020-10-10",
                            "advisor": "awesome",
                            "position": "masters researcher",
                        }
                    ],
                    "role": "masters researcher",
                    "begin_year": 2020,
                    "end_year": 2020,
                    "end_date": dt.date(2020, 10, 10),
                    "status": "ms",
                    "position": "masters researcher",
                }
            ],
        ),
        (
            [
                [
                    {
                        "_id": "student",
                        "name": "Lancelot",
                        "employment": [
                            {
                                "status": "ms",
                                "advisor": "awesome",
                                "begin_date": "2020-05-05",
                                "position": "masters researcher",
                            }
                        ],
                    }
                ],
                "2019-01-01",
            ],
            [
                {
                    "_id": "student",
                    "name": "Lancelot",
                    "employment": [
                        {
                            "status": "ms",
                            "advisor": "awesome",
                            "begin_date": "2020-05-05",
                            "position": "masters researcher",
                        }
                    ],
                    "role": "masters researcher",
                    "begin_year": 2020,
                    "end_year": "present",
                    "end_date": dt.date(2021, 6, 3),
                    "status": "ms",
                    "position": "masters researcher",
                }
            ],
        ),
    ],
)
def test_filter_employment_for_advisees(inp, expected):
    actual = filter_employment_for_advisees(inp[0], inp[1], "ms", "awesome", dt.date(2021, 6, 3))
    assert actual == expected


person1 = {"_id": "tstark", "aka": "iron man", "name": "tony stark"}
person2 = {"_id": "nromanov", "aka": "black widow", "name": "natasha romanov"}
PEOPLE = [person1, person2]

presentation1 = {
    "_id": "abc",
    "authors": "tstark",
    "date": "2018-01-01",
    "department": "apam",
    "institution": "columbiau",
    "status": "accepted",
    "type": "award",
}
presentation2 = {
    "_id": "ghi",
    "authors": ["tstark", "nromanov"],
    "begin_date": "2019-01-02",
    "end_date": "2019-01-08",
    "department": "physics",
    "institution": "rutgersu",
    "status": "cancelled",
    "type": "poster",
}
presentation3 = {
    "_id": "jkl",
    "authors": ["nromanov"],
    "begin_year": 2020,
    "begin_month": 2,
    "begin_day": 2,
    "end_year": 2020,
    "end_month": 12,
    "end_day": 12,
    "department": "math",
    "institution": "rutgersu",
    "status": "declined",
    "type": "webinar",
}
PRESENTATIONS = [presentation1, presentation2, presentation3]

institution1 = {
    "_id": "columbiau",
    "city": "New York",
    "country": "USA",
    "name": "Columbia University",
    "state": "NY",
}
institution2 = {
    "_id": "rutgersu",
    "city": "New Brunswick",
    "country": "USA",
    "name": "Rutgers University",
    "state": "NJ",
}
institution3 = {
    "_id": "barnardc",
    "city": "New York",
    "country": "USA",
    "name": "Barnard College",
    "state": "NY",
    "departments": {"physics": {"name": "Department of Physics", "aka": "Phys"}},
}
institution4 = {
    "_id": "nyu",
    "city": "New York",
    "country": "USA",
    "name": "New York University",
    "state": "NY",
    "street": "23rd",
    "zip": "10001",
    "aka": "purple",
}
institution_overseas = {"_id": "overseasu", "city": "Toronto", "country": "Canada", "name": "Overseas University"}
organization1 = {"_id": "3m", "city": "Minneapolis", "country": "USA", "name": "3M", "state": "MN"}
INSTITUTIONS = [institution1, institution2, institution3, institution4, institution_overseas, organization1]


expected1 = {
    "_id": "abc",
    "authors": "tony stark",
    "begin_day_suffix": "st",
    "begin_year": 2018,
    "begin_month": 1,
    "begin_day": 1,
    "date": dt.date(2018, 1, 1),
    "day_suffix": "st",
    "department": {"name": "apam"},
    "institution": {"city": "New York", "country": "USA", "name": "Columbia University", "state": "NY"},
    "status": "accepted",
    "type": "award",
}
expected2 = {
    "_id": "ghi",
    "authors": "tony stark, natasha romanov",
    "begin_date": "2019-01-02",
    "begin_day_suffix": "nd",
    "begin_year": 2019,
    "begin_month": 1,
    "begin_day": 2,
    "date": dt.date(2019, 1, 2),
    "day_suffix": "nd",
    "end_day": 8,
    "end_day_suffix": "th",
    "department": {"name": "physics"},
    "end_date": "2019-01-08",
    "institution": {"city": "New Brunswick", "country": "USA", "name": "Rutgers University", "state": "NJ"},
    "status": "cancelled",
    "type": "poster",
}
expected3 = {
    "_id": "jkl",
    "authors": "natasha romanov",
    "begin_day": 2,
    "begin_day_suffix": "nd",
    "begin_month": 2,
    "begin_year": 2020,
    "date": dt.date(2020, 2, 2),
    "day_suffix": "nd",
    "department": {"name": "math"},
    "end_day": 12,
    "end_day_suffix": "th",
    "end_month": 12,
    "end_year": 2020,
    "institution": {"city": "New Brunswick", "country": "USA", "name": "Rutgers University", "state": "NJ"},
    "status": "declined",
    "type": "webinar",
}


@pytest.mark.parametrize(
    "input, expected, sysout",
    [
        (
            {"institution": "columbiau"},
            {
                "department": "unknown",
                "location": "New York, NY",
                "city": "New York",
                "country": "USA",
                "institution": "Columbia University",
                "organization": "Columbia University",
                "state": "NY",
            },
            "",
        ),
        (
            {"institution": "nyu"},
            {
                "department": "unknown",
                "location": "New York, NY",
                "city": "New York",
                "country": "USA",
                "institution": "New York University",
                "organization": "New York University",
                "state": "NY",
                "street": "23rd",
                "zip": "10001",
                "aka": "purple",
            },
            "",
        ),
        (
            {"institution": "barnardc", "department": "physics"},
            {
                "location": "New York, NY",
                "city": "New York",
                "country": "USA",
                "institution": "Barnard College",
                "organization": "Barnard College",
                "state": "NY",
                "department": "Department of Physics",
            },
            "",
        ),
        (
            {"institution": "columbiau", "department": "physics"},
            {
                "location": "New York, NY",
                "city": "New York",
                "country": "USA",
                "institution": "Columbia University",
                "organization": "Columbia University",
                "state": "NY",
                "department": "physics",
            },
            "WARNING: no departments in columbiau. physics sought\n",
        ),
        (
            {"organization": "3m"},
            {
                "department": "unknown",
                "location": "Minneapolis, MN",
                "city": "Minneapolis",
                "country": "USA",
                "institution": "3M",
                "organization": "3M",
                "state": "MN",
            },
            "",
        ),
        (
            {"institution": "notindbu"},
            {
                "location": "unknown, unknown",
                "city": "unknown",
                "country": "unknown",
                "institution": "notindbu",
                "department": "unknown",
                "organization": "notindbu",
                "state": "unknown",
            },
            "WARNING: notindbu not found in institutions\n",
        ),
        (
            {"institution": "notindbu", "location": "Near, BY"},
            {
                "location": "Near, BY",
                "city": "unknown",
                "country": "unknown",
                "institution": "notindbu",
                "department": "unknown",
                "organization": "notindbu",
                "state": "unknown",
            },
            "WARNING: notindbu not found in institutions\n",
        ),
        (
            {"institution": "notindbu", "city": "Near", "state": "BY"},
            {
                "location": "Near, BY",
                "city": "Near",
                "country": "unknown",
                "institution": "notindbu",
                "department": "unknown",
                "organization": "notindbu",
                "state": "BY",
            },
            "WARNING: notindbu not found in institutions\n",
        ),
        (
            {"institution": "overseasu"},
            {
                "location": "Toronto, Canada",
                "city": "Toronto",
                "country": "Canada",
                "department": "unknown",
                "institution": "Overseas University",
                "organization": "Overseas University",
            },
            "",
        ),
        (
            {"degree": "phd"},
            {"degree": "phd"},
            "WARNING: no institution or organization in entry: {'degree': 'phd'}\n",
        ),
    ],
)
def test_dereference_institution(input, expected, sysout, capsys):
    dereference_institution(input, INSTITUTIONS, verbose=True)
    assert expected == input
    out, err = capsys.readouterr()
    assert sysout == out


@pytest.mark.parametrize(
    "args, kwargs, expected",
    [
        # this tests no kwargs
        ([PEOPLE, PRESENTATIONS, INSTITUTIONS, "tstark"], {}, [expected1]),
        # this tests 'statuses' kwarg
        ([PEOPLE, PRESENTATIONS, INSTITUTIONS, "tstark"], {"statuses": ["all"]}, [expected2, expected1]),
        # this tests 'statuses' and 'types' kwargs together
        ([PEOPLE, PRESENTATIONS, INSTITUTIONS, "tstark"], {"statuses": ["all"], "types": ["poster"]}, [expected2]),
        # this tests 'statuses' and 'since' kwargs together
        (
            [PEOPLE, PRESENTATIONS, INSTITUTIONS, "nromanov"],
            {"statuses": ["all"], "since": dt.date(2019, 1, 1)},
            [expected3, expected2],
        ),
        # this tests the 'statuses' and 'before' kwargs together
        (
            [PEOPLE, PRESENTATIONS, INSTITUTIONS, "tstark"],
            {"statuses": ["all"], "before": dt.date(2018, 1, 2)},
            [expected1],
        ),
    ],
)
def test_filter_presentations(args, kwargs, expected):
    actual = filter_presentations(*args, **kwargs)
    assert actual == expected


@pytest.mark.parametrize(
    "coll, expected",
    [
        ([{"_id": "id", "name": "test"}], []),
        ([{"_id": "id", "tags": ""}], []),
        ([{"_id": "id", "tags": "thing1"}], ["thing1"]),
        ([{"_id": "id", "tags": "thing2,thing1"}], ["thing1", "thing2"]),
        ([{"_id": "id", "tags": "thing2 thing1"}], ["thing1", "thing2"]),
        ([{"_id": "id", "tags": "thing2,thing1 thing3"}], ["thing1", "thing2", "thing3"]),
    ],
)
def test_get_tags(coll, expected):
    actual = get_tags(coll)
    assert actual == expected


def test_get_tags_invalid():
    coll = [{"_id": "id", "tags": ["test"]}]
    with pytest.raises(TypeError) as e_info:
        get_tags(coll)
        assert e_info == "ERROR: valid tags are comma or space separated strings of tag names"


@pytest.mark.parametrize(
    "repo_information, expected",
    [
        # good input
        (
            [
                {
                    "_id": "repo1",
                    "params": {"namespace_id": "35", "initialize_with_readme": "false", "name": "repo name "},
                    "url": "https://example.com",
                    "api_route": "/url/example",
                    "namespace_name": "talks",
                }
            ],
            {
                "_id": "repo1",
                "built_url": "https://example.com/url/example",
                "params": {"namespace_id": "35", "initialize_with_readme": "false", "name": "repo_name"},
                "url": "https://example.com",
                "api_route": "/url/example",
                "namespace_name": "talks",
            },
        ),
        ({}, False),
        ([], False),
        # multiple docs with same _id
        ([{"_id": "repo1"}, {"_id": "repo1"}], False),
        # well formulated doc, but wrong id
        (
            [
                {
                    "_id": "wrong_id",
                    "params": {"namespace_id": "35", "initialize_with_readme": "false", "name": "repo name "},
                    "url": "https://example.com/url/example",
                }
            ],
            False,
        ),
        # no params section
        (
            [
                {"_id": "repo1", "url": "https://example.com/url/example"},
            ],
            False,
        ),
        # params section, but empty
        ([{"_id": "repo1", "params": {}, "url": "https://example.com/url/example"}], False),
        # name but name empty
        (
            [
                {
                    "_id": "repo1",
                    "params": {"namespace_id": "", "initialize_with_readme": "false", "name": ""},
                    "url": "https://example.com/url/example",
                }
            ],
            False,
        ),
        # url but url empty
        (
            [
                {
                    "_id": "repo1",
                    "params": {"namespace_id": "1", "initialize_with_readme": "false", "name": "repo name"},
                    "url": "",
                }
            ],
            False,
        ),
        # url but url not complete
        (
            [
                {
                    "_id": "repo1",
                    "params": {"namespace_id": "1", "initialize_with_readme": "false", "name": "repo name"},
                    "url": "https://example.com",
                }
            ],
            False,
        ),
        # url but url invalid
        (
            [
                {
                    "_id": "repo1",
                    "params": {"name": "some name", "namespace_id": "1", "initialize_with_readme": "false"},
                    "url": "random junk",
                }
            ],
            False,
        ),
    ],
)
def test_get_target_repo_info(repo_information, expected):
    actual = get_target_repo_info("repo1", repo_information)
    assert actual == expected


@pytest.mark.parametrize(
    "tokens, expected",
    [
        ([{"_id": "gitlab_private_token", "token": "<private-token>"}], "<private-token>"),
        ([{"_id": "wrong_name", "token": "<private-token>"}], None),
        (
            [
                {"_id": "gitlab_private_token", "token": "<private-token>"},
                {"_id": "gitlab_private_token", "token": "<private-token>"},
            ],
            None,
        ),
        ([{"_id": "gitlab_private_token", "token": ""}], None),
        ([{"_id": "gitlab_private_token"}], None),
        ([{"_id": "<private-token>"}], None),
        ({}, None),
    ],
)
def test_get_target_token(tokens, expected):
    actual = get_target_token("gitlab_private_token", tokens)
    assert actual == expected


# @mock.patch("requests.post")


@requests_mock.Mocker(kw="mock")
def test_create_repo(**kwargs):
    kwargs["mock"].post("https://example.com/url/example", status_code=201)
    # mock_requests_post.return_value = mock.Mock(**{"status_code": 201})
    rc = copy.copy(DEFAULT_RC)
    repo_token_information = {
        "repos": [
            {
                "_id": "talk_repo",
                "params": {"namespace_id": "35", "initialize_with_readme": "false", "name": "2206_my_talk"},
                "url": "https://example.com",
                "api_route": "/url/example",
                "namesapce_name": "talks",
            }
        ],
        "tokens": [{"_id": "gitlab_private_token", "token": "<private-token>"}],
    }
    rc._update(repo_token_information)
    actual = create_repo("talk_repo", "gitlab_private_token", rc)
    assert (
        actual == "repo 2206_my_talk has been created at https://example.com.\nClone this to your local using "
        "(HTTPS):\ngit clone https://example.com:<group/org name>/2206_my_talk.git\nor "
        "(SSH):\ngit clone git@example.com:<group/org name>/2206_my_talk.git"
    )


# @mock.patch('uuid.uuid4', return_value="test-uid")
@pytest.fixture
def test_get_uuid(mocker):
    mocker.patch("uuid.uuid4", return_value="test-uuid")
    expected = "test-uuid"
    actual = get_uuid()
    assert expected == actual


tga = [
    (
        {
            "_id": "good_person",
            "appointments": {
                "appt_id_1": {
                    "begin_date": "2022-06-01",
                    "end_date": "2022-06-30",
                    "grant": "good_grant",
                    "loading": 0.5,
                    "status": "submitted",
                    "type": "gra",
                    "notes": "",
                }
            },
        },
        [],
        "good_grant",
        [("good_person", dt.date(2022, 6, 1), dt.date(2022, 6, 30), 0.5, 0.48, "good_grant")],
    ),
    (
        {
            "_id": "good_person",
            "appointments": {
                "appt_id_1": {
                    "begin_date": "2022-06-01",
                    "end_date": "2022-06-30",
                    "grant": "good_grant",
                    "loading": 0.5,
                    "status": "submitted",
                    "type": "gra",
                    "notes": "",
                }
            },
        },
        [],
        None,
        [("good_person", dt.date(2022, 6, 1), dt.date(2022, 6, 30), 0.5, 0.48, "good_grant")],
    ),
    (
        {
            "_id": "good_person",
            "appointments": {
                "appt_id_1": {
                    "begin_date": "2022-06-01",
                    "end_date": "2022-06-30",
                    "grant": "good_grant",
                    "loading": 0.5,
                    "status": "submitted",
                    "type": "gra",
                    "notes": "",
                }
            },
        },
        [],
        "bad_grant",
        [],
    ),
]


@pytest.mark.parametrize("tga", tga)
def test_get_appointments(tga):
    expected = tga[3]
    actual = get_appointments(tga[0], tga[1], tga[2])
    assert expected == actual



from collections import defaultdict
from copy import deepcopy

from regolith.fsclient import FileSystemClient
from regolith.mongoclient import MongoClient

CLIENTS = {
    "mongo": MongoClient,
    "mongodb": MongoClient,
    "fs": FileSystemClient,
    "filesystem": FileSystemClient,
}


class ClientManager:
    """
    Client wrapper that allows for multiple backend clients to be used in parallel with one chained DB
    """

    def __init__(self, databases, rc):
        client_tuple = tuple()
        if hasattr(rc, "backend"):
            for database in databases:
                database["backend"] = rc.backend
        for database in databases:
            if "backend" not in database:
                database["backend"] = "filesystem"
            backend_object_type = CLIENTS[database["backend"]]
            # Checks to see if the clients tuple contains a client with the database's backend
            if len(client_tuple) == 0:
                client_tuple = client_tuple + (CLIENTS[database["backend"]](rc),)
            elif True not in [isinstance(client, backend_object_type) for client in client_tuple]:
                client_tuple = client_tuple + (CLIENTS[database["backend"]](rc),)
        self.clients = client_tuple
        self.rc = rc
        self.closed = True
        self.chained_db = None
        # self.open()
        self._collfiletypes = {}
        self._collexts = {}
        self._yamlinsts = {}

    def __getattr__(self, attr):
        if attr == "dbs":
            concatenated_dbs_dict = defaultdict(lambda: defaultdict(dict))
            for client in self.clients:
                concatenated_dbs_dict.update(client.dbs)
            return concatenated_dbs_dict
        else:
            raise AttributeError

    def __getitem__(self, key):
        for client in self.clients:
            if key in client.keys():
                return client[key]

    def open(self):
        """Opens the database connections"""
        for client in self.clients:
            client.open()

    def close(self):
        """Closes the database connections."""
        for client in self.clients:
            client.close()

    def load_database(self, db):
        for client in self.clients:
            if isinstance(client, CLIENTS[db["backend"]]):
                client.load_database(db)

    def import_database(self, db: dict):
        for client in self.clients:
            if isinstance(client, MongoClient):
                client.import_database(db)

    def export_database(self, db: dict):
        for client in self.clients:
            if isinstance(client, MongoClient):
                client.export_database(db)

    def dump_database(self, db):
        to_add = []
        # Iterate through the clients just in case databases on different backends have same name
        for client in self.clients:
            if isinstance(client, CLIENTS[db["backend"]]):
                if db["name"] in client.keys():
                    temp_add = client.dump_database(db)
                    if temp_add:
                        to_add.extend(temp_add)
        return to_add

    def keys(self):
        keys = []
        for client in self.clients:
            keys.append(client.keys())
        return keys

    def collection_names(self, dbname, include_system_collections=True):
        """Returns the collaction names for a database."""
        for client in self.clients:
            if dbname in client.keys():
                return client.collection_names(dbname)

    def all_documents(self, collname, copy=True):
        """Returns an iteratable over all documents in a collection."""
        if copy:
            return deepcopy(self.chained_db.get(collname, {})).values()
        return self.chained_db.get(collname, {}).values()

    def insert_one(self, dbname, collname, doc):
        """Inserts one document to a database/collection."""
        for client in self.clients:
            if dbname in client.keys():
                client.insert_one(dbname, collname, doc)

    def insert_many(self, dbname, collname, docs):
        """Inserts many documents into a database/collection."""
        for client in self.clients:
            if dbname in client.keys():
                client.insert_many(dbname, collname, docs)

    def delete_one(self, dbname, collname, doc):
        """Removes a single document from a collection"""
        for client in self.clients:
            if dbname in client.keys():
                client.delete_one(dbname, collname, doc)

    def find_one(self, dbname, collname, filter):
        """Finds the first document matching filter."""
        for client in self.clients:
            if dbname in client.keys():
                return client.find_one(dbname, collname, filter)

    def update_one(self, dbname, collname, filter, update, **kwargs):
        """Updates one document."""
        for client in self.clients:
            if dbname in client.keys():
                client.update_one(dbname, collname, filter, update, **kwargs)



"""Date based tools"""

import calendar
import datetime

from dateutil import parser as date_parser

MONTHS = {
    "jan": 1,
    "jan.": 1,
    "january": 1,
    "feb": 2,
    "feb.": 2,
    "february": 2,
    "mar": 3,
    "mar.": 3,
    "march": 3,
    "apr": 4,
    "apr.": 4,
    "april": 4,
    "may": 5,
    "may.": 5,
    "jun": 6,
    "jun.": 6,
    "june": 6,
    "jul": 7,
    "jul.": 7,
    "july": 7,
    "aug": 8,
    "aug.": 8,
    "august": 8,
    "sep": 9,
    "sep.": 9,
    "sept": 9,
    "sept.": 9,
    "september": 9,
    "oct": 10,
    "oct.": 10,
    "october": 10,
    "nov": 11,
    "nov.": 11,
    "november": 11,
    "dec": 12,
    "dec.": 12,
    "december": 12,
    "": 1,
    "tbd": 12,
}


def month_to_int(m):
    """Converts a month to an integer."""
    try:
        m = int(m)
    except ValueError:
        m = MONTHS[m.lower()]
    return m


def month_to_str_int(m):
    """Converts a month to an int form, str type, with a leading zero"""
    mi = month_to_int(m)
    if mi < 10:
        ms = "0{}".format(mi)
    else:
        ms = str(mi)
    return ms


def day_to_str_int(d):
    """Converts a day to an int form, str type, with a leading zero"""
    if d < 10:
        ds = "0{}".format(d)
    else:
        ds = str(d)
    return ds


def date_to_float(y, m, d=0):
    """Converts years / months / days to a float, eg 2015.0818 is August
    18th 2015."""
    y = int(y)
    m = month_to_int(m)
    d = int(d)
    return y + (m / 100.0) + (d / 10000.0)


def find_gaps_overlaps(dateslist, overlaps_ok=False):
    """
    Find whether there is a gap or an overlap in a list of date-ranges

    Parameters
    ----------
    dateslist: list of tuples of datetime.date objects
      The list of date-ranges.
    overlaps_ok: bool
      Returns false if there are gaps but true if there are overlaps but no gaps

    Returns
    -------
      True if there are no gaps or overlaps else False

    """

    status = True
    dateslist.sort(key=lambda x: x[0])
    for i in range(len(dateslist) - 1):
        if dateslist[i + 1][0] <= dateslist[i][1] and not overlaps_ok:
            status = False
        elif (dateslist[i + 1][0] - dateslist[i][1]).days > 1:
            status = False
    return status


def last_day(year, month):
    """
    Returns the last day of the month for the month given

    Parameters
    ----------
    year: integer
      the year that the month is in
    month: integer or string
      the month.  if a string should be resolvable using regolith month_to_int

    Returns
    -------
    The last day of that month

    """
    return calendar.monthrange(year, month_to_int(month))[1]


def convert_doc_iso_to_date(doc):
    def convert_date(obj):
        """
        Recursively goes through the dictionary obj and converts date from iso to datetime.date
        """
        if isinstance(obj, str):
            try:
                date = datetime.datetime.strptime(obj, "%Y-%m-%d").date()
            except ValueError:
                return obj
            else:
                return date
        if isinstance(obj, (int, float)):
            return obj
        if isinstance(obj, dict):
            new = obj.__class__()
            for k, v in obj.items():
                new[k] = convert_date(v)
        elif isinstance(obj, (list, set, tuple)):
            new = obj.__class__(convert_date(v) for v in obj)
        else:
            return obj
        return new

    return convert_date(doc)


def get_dates(thing, date_field_prefix=None):
    """
    given a dict like thing, return the date items

    Parameters
    ----------
    thing: dict
      the dict that contains the dates
    date_field_prefix: string (optional)
      the prefix to look for before the date parameter. For example given "submission"
      the function will search for submission_day, submission_year, etc.

    Returns
    -------
       dict containing datetime.date objects for valid begin_date end_date and date, and
       prefix_date if a prefix string was passed.  Missing and empty dates and
       date items that contain the string 'tbd' are not returned.  If no valid
       date items are found, an empty dict is returned

    Description
    -----------
    If "begin_date", "end_date" or "date" values are found, if these are are in
    an ISO format string they will be converted to datetime.date objects and
    returned in the dictionary under keys of the same name.  A specified date
    will override any date built from year/month/day data.

    If they are not found the function will look for begin_year, end_year and
    year.

    If "year", "month" and "day" are found the function will return these in the
    "date" field and begin_date and end_date will match the "date" field. If only
    a "year" is found, then the date attribute will be none but the begin and end
    dates will be the first and last day of that respective year.

    If year is found but no month or day are found the function will return
    begin_date and end_date with the beginning and the end of the given year/month.
    The returned date will be None.

    If end_year is found, the end month and end day are missing they are set to
    12 and 31, respectively

    If begin_year is found, the begin month and begin day are missing they are set to
    1 and 1, respectively

    If a date field prefix is passed in this function will search for prefix_year as
    well as prefix_month, prefix_day, and prefix_date. For example, if the prefix string
    passed in is "submitted" then this function will look for submitted_date instead of
    just date.

    Examples
    --------
    >>> get_dates({'submission_day': 10, 'submission_year': 2020, 'submission_month': 'Feb'}, "submission")

    This would return a dictionary consisting of the begin date, end, date, and date for the given input.
    Instead of searching for "day" in the thing, it would search for "submission_day" since a prefix was
    given. The following dictionary is returned (note that a submission_date and a date key are in the
    dictionary):
    {'begin_date': datetime.date(2020, 2, 10),
     'end_date': datetime.date(2020, 2, 10),
     'submission_date': datetime.date(2020, 2, 10),
     'date': datetime.date(2020, 2, 10)
    }

    >>> get_dates({'begin_year': 2019, 'end_year': 2020, 'end_month': 'Feb'})

    This will return a dictionary consisting of the begin date, end date, and date for the given input.
    Because no prefix string was passed in, the function will search for "date" in the input instead of
    prefix_input. The following dictionary is returned:
    {'begin_date': datetime.date(2019, 1, 1),
     'end_date': datetime.date(2020, 2, 29),
    }
    """

    datenames = ["day", "month", "year", "date"]
    if date_field_prefix:
        datenames = [f"{date_field_prefix}_{datename}" for datename in datenames]

    minimal_set = ["end_year", "begin_year", "year", "begin_date", "end_date", "date"]
    minimal_things = (
        list(set([thing.get(i) for i in datenames]))
        if date_field_prefix
        else list(set([thing.get(i) for i in minimal_set]))
    )
    if len(minimal_things) == 1 and not minimal_things[0]:
        print(f"WARNING: cannot find any dates in {thing.get('_id', '(no id)')}")
        dates = {}
        return dates
    for key, value in thing.items():
        if key in minimal_set or key in ["month", "day", "begin_day", "begin_month"]:
            if isinstance(value, str):
                if value.strip().lower() == "tbd":
                    thing[key] = None
                else:
                    try:
                        thing[key] = int(value)
                    except ValueError:
                        pass
    if thing.get("end_year") and not thing.get("begin_year"):
        print(f"WARNING: end_year specified without begin_year {thing.get('_id', '(no id)')}")
    begin_date, end_date, date = None, None, None
    if thing.get("begin_year"):
        if not thing.get("begin_month"):
            thing["begin_month"] = 1
        if not thing.get("begin_day"):
            thing["begin_day"] = 1
        begin_date = datetime.date(thing["begin_year"], month_to_int(thing["begin_month"]), thing["begin_day"])
    if thing.get("end_year"):
        if not thing.get("end_month"):
            thing["end_month"] = 12
        if not thing.get("end_day"):
            thing["end_day"] = last_day(thing["end_year"], thing["end_month"])
        end_date = datetime.date(thing["end_year"], month_to_int(thing["end_month"]), thing["end_day"])
    if thing.get(datenames[2]):  # prefix_year
        if not thing.get(datenames[1]):  # prefix_month
            if thing.get("begin_year"):
                print(
                    f"WARNING: both year and begin_year specified in {thing.get('_id', '(no id)')}. "
                    "Year info will be used"
                )
            begin_date = datetime.date(thing[datenames[2]], 1, 1)
            end_date = datetime.date(thing[datenames[2]], 12, 31)
        elif not thing.get(datenames[0]):  # prfix_day
            if thing.get("begin_year"):
                print(
                    f"WARNING: both year and begin_year specified in {thing.get('_id', '(no id)')}. "
                    "Year info will be used"
                )
            begin_date = datetime.date(thing[datenames[2]], month_to_int(thing[datenames[1]]), 1)
            end_date = datetime.date(
                thing[datenames[2]],
                month_to_int(thing[datenames[1]]),
                last_day(thing[datenames[2]], thing[datenames[1]]),
            )
        else:
            date = datetime.date(thing[datenames[2]], month_to_int(thing[datenames[1]]), int(thing[datenames[0]]))
            begin_date = datetime.date(
                int(thing[datenames[2]]), month_to_int(thing[datenames[1]]), int(thing[datenames[0]])
            )
            end_date = datetime.date(
                int(thing[datenames[2]]), month_to_int(thing[datenames[1]]), int(thing[datenames[0]])
            )
    if thing.get("begin_date"):
        if isinstance(thing.get("begin_date"), str):
            begin_date = date_parser.parse(thing.get("begin_date")).date()
        else:
            begin_date = thing.get("begin_date")
    if thing.get("end_date"):
        if isinstance(thing.get("end_date"), str):
            end_date = date_parser.parse(thing.get("end_date")).date()
        else:
            end_date = thing.get("end_date")
    if thing.get(datenames[3]):
        if isinstance(thing.get(datenames[3]), str):
            date = date_parser.parse(thing.get(datenames[3])).date()
        else:
            date = thing.get(datenames[3])

    if date_field_prefix:
        dates = {"begin_date": begin_date, "end_date": end_date, datenames[3]: date, "date": date}
    else:
        dates = {"begin_date": begin_date, "end_date": end_date, "date": date}
    dates_no_nones = {k: v for k, v in dates.items() if v is not None}
    return dates_no_nones


def get_due_date(thing):
    """

    Parameters
    ----------
    thing: dict
      gets the field named 'due_date' from doc and ensurese it is a
      datetime.date object

    Returns
    -------
    The due date as a datetime.date object

    """
    due_date = thing.get("due_date")
    if isinstance(due_date, str):
        due_date = date_parser.parse(due_date).date()
    elif isinstance(due_date, datetime.date):
        pass
    else:
        raise RuntimeError("due date not a known type")
    return due_date


def is_current(thing, now=None):
    """
    given a thing with dates, returns true if the thing is current
    looks for begin_ and end_ daty things (date, year, month, day), or just
    the daty things themselves. e.g., begin_date, end_month, month, and so on.

    Parameters
    ----------
    thing: dict
      the thing that we want to know whether or not it is current
    now: datetime.date object
      a date for now.  If it is None it uses the current date.  Default is None

    Returns
    -------
    True if the thing is current and false otherwise

    """
    if now is None:
        now = datetime.date.today()
    dates = get_dates(thing)
    current = False
    if not dates.get("end_date"):
        dates["end_date"] = datetime.date(5000, 12, 31)
    try:
        if dates.get("begin_date") <= now <= dates.get("end_date"):
            current = True
    except RuntimeError:
        raise RuntimeError(f"Cannot find begin_date in document:\n {thing['_id']}")
    return current


def has_started(thing, now=None):
    """
    given a thing with dates, returns true if the thing has started

    Parameters
    ----------
    thing: dict
      the thing that we want to know whether or not it is has started
    now: datetime.date object
      a date for now.  If it is None it uses the current date.  Default is None

    Returns
    -------
    True if the thing has started and false otherwise

    """
    if not now:
        now = datetime.date.today()
    dates = get_dates(thing)
    started = False
    try:
        if dates.get("begin_date") <= now:
            started = True
    except RuntimeError:
        raise RuntimeError(f"Cannot find begin_date in document:\n {thing}")
    return started


def has_finished(thing, now=None):
    """
    given a thing with dates, returns true if the thing has finished

    Parameters
    ----------
    thing: dict
      the thing that we want to know whether or not it has finished
    now: datetime.date object
      a date for now.  If it is None it uses the current date.  Default is None

    Returns
    -------
    True if the thing has finished and false otherwise

    """
    if not now:
        now = datetime.date.today()
    dates = get_dates(thing)
    finished = False
    if not dates.get("end_date"):
        dates["end_date"] = datetime.date(5000, 12, 31)
    if dates.get("end_date") < now:
        finished = True
    return finished


def is_before(thing, now=None):
    """
    given a thing with a date, returns true if the thing is before the input date

    Parameters
    ----------
    thing: dict
      the thing that we want to know whether or not is before a date
    now: datetime.date object
      a date for now.  If it is None it uses the current date.  Default is None

    Returns
    -------
    True if the thing is before the date

    """
    if not now:
        now = datetime.date.today()
    dates = get_dates(thing)
    before = False
    try:
        if dates.get("date") < now:
            before = True
    except RuntimeError:
        raise RuntimeError(f"Cannot find date in document:\n {thing}")
    return before


def is_after(thing, now=None):
    """
    given a thing with a date, returns true if the thing is after the input date

    Parameters
    ----------
    thing: dict
      the thing that we want to know whether or not is after a date
    now: datetime.date object
      a date for now.  If it is None it uses the current date.  Default is None

    Returns
    -------
    True if the thing is after the date

    """
    if not now:
        now = datetime.date.today()
    dates = get_dates(thing)
    after = False
    try:
        if now < dates.get("date"):
            after = True
    except RuntimeError:
        raise RuntimeError(f"Cannot find date in document:\n {thing}")
    return after


def is_between(thing, start=None, end=None):
    """
    given a thing with a date, returns true if the thing is between the start and end date

    Parameters
    ----------
    thing: dict
      the thing that we want to know whether or not is after a date
    start: datetime.date object
      a date for the start.  If it is None it uses the current date.  Default is None
    end: datetime.date object
      a date for the end.  If it is None it uses the current date.  Default is None

    Returns
    -------
    True if the thing is between the start and end

    """
    if not start:
        start = datetime.date.today()
    if not end:
        end = datetime.date.today()
    between = False
    if is_after(thing, start) and is_before(thing, end):
        between = True
    return between



"""Helper for adding a new person to the contacts collection."""

import datetime as dt
import uuid

import dateutil.parser as date_parser
from gooey import GooeyParser
from nameparser import HumanName

from regolith.fsclient import _id_key
from regolith.helpers.basehelper import DbHelperBase
from regolith.tools import all_docs_from_collection, fragment_retrieval

TARGET_COLL = "contacts"


def subparser(subpi):
    date_kwargs = {}
    notes_kwargs = {}
    if isinstance(subpi, GooeyParser):
        date_kwargs["widget"] = "DateChooser"
        notes_kwargs["widget"] = "Textarea"

    subpi.add_argument("fragmentname", help="Fragment of name, id, or aliases to search for contacts.")
    subpi.add_argument(
        "-i", "--index", help="Index of the item in the enumerated list chosen to update.", type=int
    )
    subpi.add_argument("-n", "--name", help="Full name. Required if new contact.")
    subpi.add_argument(
        "-o",
        "--institution",
        help="Person's institution. It can be "
        "institution id or anything in the "
        "aka or name from institutions collection. "
        "It is required to create a new contact.",
    )
    subpi.add_argument(
        "-t",
        "--notes",
        nargs="+",
        help="Notes.  As many notes as you like, each one in "
        "quotes and separated by a space, such as where "
        "and when met, what discussed.",
        **notes_kwargs,
    )
    subpi.add_argument("-d", "--department", help="Department at the institution.")
    subpi.add_argument(
        "--id", help="id of the person, e.g., first letter first name " "plus last name, but unique."
    )
    subpi.add_argument(
        "--aliases",
        nargs="+",
        help="All the different ways that the person may "
        "be referred to as.  As many as you like, in "
        "quotes separated by a space",
    )
    # Do not delete --date arg
    subpi.add_argument(
        "--date", help="The date when the contact was created. " "Defaults to today's date.", **date_kwargs
    )
    # Do not delete --database arg
    subpi.add_argument(
        "--database",
        help="The database that will be updated.  Defaults to " "first database in the regolithrc.json file.",
    )
    # FIXME
    # subpi.add_argument("-e", "--email",
    #                    help="email address")

    return subpi


class ContactUpdaterHelper(DbHelperBase):
    """Helper for adding a new person to the contacts collection."""

    # btype must be the same as helper target in helper.py
    btype = "u_contact"
    needed_colls = [f"{TARGET_COLL}"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        rc.coll = f"{TARGET_COLL}"
        if not rc.database:
            rc.database = rc.databases[0]["name"]
        gtx[rc.coll] = sorted(all_docs_from_collection(rc.client, rc.coll), key=_id_key)
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def db_updater(self):
        rc = self.rc
        found_contacts = fragment_retrieval(self.gtx["contacts"], ["_id", "aka", "name"], rc.fragmentname)
        found_contacts.sort(key=lambda x: x["_id"], reverse=False)
        index_list = list(range(2, (len(found_contacts) + 2)))
        if not rc.index:
            print(
                "Please rerun the helper by hitting up arrow and adding '-i list-index' "
                "to update the list item 'list-index', e.g., 'regolith helper eins -i 2'. "
                "For new contacts --name (-n) and --institution (-o) are required:"
            )
            print(f"{1}. {rc.fragmentname} as a new contact")
            for i, j in zip(index_list, found_contacts):
                print(
                    f"{i}. {j.get('name')}\n"
                    f"   id: {j.get('_id')}\n"
                    f"   email: {j.get('email')}\n"
                    f"   institution: {j.get('institution')}\n"
                    f"   department: {j.get('department')}\n"
                    f"   notes: {j.get('notes')}\n"
                    f"   aliases: {j.get('aka')}"
                )
            return
        pdoc = {}
        if int(rc.index) == 1:
            if not rc.institution or not rc.name:
                raise RuntimeError("Institution and name are required to create a new contact")
            name = HumanName(rc.name)
            if not rc.id:
                key = str(name.first[0].lower().replace(" ", "") + name.last.lower().replace(" ", ""))
            else:
                key = rc.id
            pdoc.update({"name": name.full_name})
            pdoc.update({"date": dt.date.today()})
            pdoc.update({"institution": rc.institution})
            notes = []
            aliases = []
            uniqueidentifier = str(uuid.uuid4())
            pdoc.update({"uuid": uniqueidentifier})
        else:
            current = found_contacts[rc.index - 2]
            key = current.get("_id")
            notes = current.get("notes", [])
            aliases = current.get("aka", [])
        if not rc.date:
            now = dt.datetime.now()
        else:
            now = date_parser.parse(rc.date).date()
        if rc.aliases:
            aliases.extend(rc.aliases)
        if rc.notes:
            if isinstance(rc.notes, str):
                rc.notes.list()
            notes.extend(rc.notes)
        if rc.department:
            pdoc.update({"department": rc.department})
        pdoc.update({"aka": aliases})
        pdoc.update({"notes": notes})
        pdoc.update({"updated": now})
        rc.client.update_one(rc.database, rc.coll, {"_id": key}, pdoc)
        print("{} has been added/updated in contacts".format(key))

        return



"""Helper for listing upcoming (and past) projectum milestones.

Projecta are small bite-sized project quanta that typically will result in
one manuscript.
"""

from gooey import GooeyParser

from regolith.dates import get_due_date
from regolith.fsclient import _id_key
from regolith.helpers.basehelper import SoutHelperBase
from regolith.schemas import (
    PROJECTUM_ACTIVE_STATI,
    PROJECTUM_CANCELLED_STATI,
    PROJECTUM_FINISHED_STATI,
    PROJECTUM_PAUSED_STATI,
    alloweds,
)
from regolith.tools import all_docs_from_collection, collection_str, get_pi_id, key_value_pair_filter

PROJECTUM_STATI = alloweds.get("PROJECTUM_STATI")
TARGET_COLL = "projecta"
HELPER_TARGET = "l_milestones"
PROJECTUM_STATI.append("all")
INACTIVE_STATI = PROJECTUM_PAUSED_STATI + PROJECTUM_CANCELLED_STATI + PROJECTUM_FINISHED_STATI
ROLES = ["pi", "lead", "group_members", "collaborators"]


def subparser(subpi):
    listbox_kwargs = {}
    if isinstance(subpi, GooeyParser):
        listbox_kwargs["widget"] = "Listbox"

    subpi.add_argument(
        "--helper-help",
        action="store_true",
        help=f"This helper will list all ACTIVE milestones by default. To "
        f"be active the projectum itself must be active and the "
        f"milestone within the prum also active.  Active milestones "
        f"have a status from {*PROJECTUM_ACTIVE_STATI, }.\n "
        f"Please erun specifying "
        f"--lead PERSON or --person PERSON. "
        f"Rerun with --all, --current, or --finished to get "
        f"all, active or finished milestones, respectively. "
        f"filters work with AND logic so -l PERSON -c will list "
        f"the active milestones for projecta where PERSON is lead. "
        f'Note that "checklist" prums are not listed. Please use '
        f"u_milestone to see those.",
    )
    subpi.add_argument(
        "-l", "--lead", help="Filter milestones for this project lead specified as an ID, " "e.g., sbillinge"
    )
    subpi.add_argument(
        "-p",
        "--person",
        help="Filter milestones for this person whether lead or not, specified " "by id, e.g., sbillinge",
    )
    subpi.add_argument(
        "-c", "--current", action="store_true", help="Same behavior as default.  Here for consistency"
    )
    subpi.add_argument(
        "--finished",
        action="store_true",
        help=f"Lists all finished milestones, i.e., status is in {*PROJECTUM_FINISHED_STATI, }",
    )
    subpi.add_argument(
        "--by-prum", action="store_true", help="Valid milestones are listed in time-order but grouped by prum"
    )
    subpi.add_argument("-v", "--verbose", action="store_true", help="increase verbosity of output")
    subpi.add_argument(
        "-s",
        "--stati",
        nargs="+",
        choices=PROJECTUM_STATI,
        help=f"Filter milestones for these stati from {*PROJECTUM_STATI, }. "
        f"Default is active projecta, i.e. {*PROJECTUM_ACTIVE_STATI, }",
        default=None,
        **listbox_kwargs,
    )
    subpi.add_argument("--all", action="store_true", help="Lists all milestones in general")
    # The --filter and --keys flags should be in every lister
    subpi.add_argument("-f", "--filter", nargs="+", help="Search this collection by giving key element pairs")
    subpi.add_argument(
        "-k",
        "--keys",
        nargs="+",
        help="Specify what keys to return values from when running "
        "--filter. If no argument is given the default is just the id.",
    )
    return subpi


class MilestonesListerHelper(SoutHelperBase):
    """Helper for listing upcoming (and past) projectum milestones.

    Projecta are small bite-sized project quanta that typically will result in
    one manuscript.
    """

    # btype must be the same as helper target in helper.py
    btype = HELPER_TARGET
    needed_colls = [f"{TARGET_COLL}"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        if "groups" in self.needed_colls:
            rc.pi_id = get_pi_id(rc)
        rc.coll = f"{TARGET_COLL}"
        colls = [
            sorted(all_docs_from_collection(rc.client, collname), key=_id_key) for collname in self.needed_colls
        ]
        for db, coll in zip(self.needed_colls, colls):
            gtx[db] = coll
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def sout(self):
        rc = self.rc
        # This if statement should be in all listers.
        # Make sure to change self.gtx to get the database the lister needs
        if rc.filter:
            collection = key_value_pair_filter(self.gtx["projecta"], rc.filter)
        else:
            collection = self.gtx["projecta"]

        if (
            (not rc.lead)
            and (not rc.verbose)
            and (not rc.stati)
            and (not rc.current)
            and (not rc.person)
            and (not rc.all)
        ):
            return

        # fixme there must be a better way, but for now use magic to
        # to remove checklists
        # print(collection)
        collection = [
            prum for prum in collection if "checklist" not in prum.get("deliverable", {}).get("scope", [])
        ]

        if rc.lead:
            if rc.person:
                raise RuntimeError("please specify either lead or person, not both")
            collection = [prum for prum in collection if prum.get("lead") == rc.lead]
        if rc.person:
            if isinstance(rc.person, str):
                rc.person = [rc.person]
            collection = [
                prum
                for prum in collection
                if prum.get("lead") in rc.person
                or bool(set(prum.get("group_members", [])).intersection(set(rc.person)))
                or bool(set(prum.get("collaborators", [])).intersection(set(rc.person)))
            ]

        if not rc.all:
            collection = [prum for prum in collection if prum.get("status") not in INACTIVE_STATI]

        all_milestones = []
        if not rc.stati and not rc.all and not rc.finished:
            rc.stati = PROJECTUM_ACTIVE_STATI
        elif rc.finished:
            rc.stati = PROJECTUM_FINISHED_STATI
        elif rc.all:
            rc.stati = PROJECTUM_STATI
        for projectum in collection:
            projectum["deliverable"].update(
                {"name": "deliverable", "objective": "deliver", "uuid": projectum.get("_id")}
            )
            milestones = [projectum["deliverable"]]
            if projectum.get("kickoff"):
                projectum["kickoff"].update({"type": "meeting", "uuid": f"ko{projectum.get('_id')}"})
                milestones = [projectum["kickoff"], projectum["deliverable"]]
            milestones.extend(projectum["milestones"])
            milestones = [ms for ms in milestones if ms.get("status") in rc.stati]

            for ms in milestones:
                due_date = get_due_date(ms)
                ms.update(
                    {
                        "lead": projectum.get("lead"),
                        "group_members": projectum.get("group_members"),
                        "collaborators": projectum.get("collaborators"),
                        "id": projectum.get("_id"),
                        "due_date": due_date,
                        "log_url": projectum.get("log_url"),
                        "pi": projectum.get("pi_id"),
                    }
                )
            milestones.sort(key=lambda x: x["due_date"], reverse=True)
            all_milestones.extend(milestones)
        if rc.keys:
            results = collection_str(all_milestones, rc.keys)
            print(results, end="")
            return
        if not rc.by_prum:
            all_milestones.sort(key=lambda x: x["due_date"], reverse=True)
        prum = ""
        for ms in all_milestones:
            if rc.by_prum:
                if prum != ms.get("id"):
                    print("-" * 50)
                    prum = ms.get("id")
            if rc.verbose:
                print(
                    f"{ms.get('due_date')} ({(ms.get('uuid')[:6])}): lead: {ms.get('lead')}, {ms.get('id')}, "
                    f"status: {ms.get('status')}"
                )
                print(f"    Type: {ms.get('type', '')}")
                print(f"    Title: {ms.get('name')}")
                print(f"    log url: {ms.get('log_url')}")
                print(f"    Purpose: {ms.get('objective')}")
                audience = []
                for i in ms.get("audience", []):
                    if isinstance(ms.get(i, i), str):
                        audience.append(ms.get(i, i))
                    else:
                        if ms.get(i):
                            audience.extend(ms.get(i))
                out = ", ".join(audience)
                print(f"    Audience: {out}")
                if ms.get("notes"):
                    print("    Notes:")
                    for note in ms.get("notes"):
                        print(f"      - {note}")
                print(f"    uuid: {ms.get('uuid')}")
            else:
                print(
                    f"{ms.get('due_date')} ({(ms.get('uuid')[:6])}): lead: {ms.get('lead')}, {ms.get('id')}, "
                    f"{ms.get('name')}, status: {ms.get('status')}"
                )

        return



"""Builder for Recent Collaborators.

For the specified person it returns the name and institution of all
graduate student and post-doc advisors,
all graduate student advisees, all post-doc advisees in the past 60 monts
and all coauthors in the past 48 months.
"""

import datetime as dt
import os
import sys
from copy import copy

import openpyxl
from dateutil.relativedelta import relativedelta
from nameparser import HumanName

from regolith.builders.basebuilder import BuilderBase
from regolith.dates import get_dates, is_after, month_to_int
from regolith.sorters import position_key
from regolith.tools import all_docs_from_collection, filter_publications, fuzzy_retrieval

NUM_COAUTHOR_MONTHS = 48
NUM_POSTDOC_MONTHS = None


def get_advisors_name_inst(advisee, rc):
    """Get the advisee's advisor. Yield (last name, first name, institution name)."""

    phd_advisors = [
        {
            "name": i.get("advisor", "missing name"),
            "type": "advisor",
            "advis_type": "phd",
            "interaction_date": get_dates(i).get("end_date", get_dates(i).get("begin_date")),
        }
        for i in advisee.get("education", [])
        if "phd" in i.get("degree", "").lower() or "dphil" in i.get("degree", "").lower()
    ]
    pdoc_advisors = [
        {
            "name": i.get("advisor", "missing name"),
            "type": "advisor",
            "advis_type": "postdoc",
            "interaction_date": get_dates(i).get("end_date", get_dates(i).get("begin_date")),
        }
        for i in advisee.get("employment", [])
        if i.get("status") == "postdoc"
    ]
    advisors = phd_advisors + pdoc_advisors
    return retrieve_names_and_insts(rc, advisors)


def get_advisees_name_inst(coll, advisor, rc):
    """Get advisor's advisees. Yield (last name, first name, institutions)"""
    advisor_names = advisor.get("aka", []) + [advisor.get("name"), advisor.get("_id")]
    advisees = []
    for person in coll:
        my_eme = person.get("employment", []) + person.get("education", [])
        relevant_emes = [i for i in my_eme if i.get("advisor", "") in advisor_names]
        phd_advisees = [
            {
                "name": person.get("name", "missing name"),
                "type": "advisee",
                "interaction_date": get_dates(i).get("end_date", get_dates(i).get("date", dt.date.today())),
                "advis_type": "phd",
            }
            for i in relevant_emes
            if "phd" in i.get("degree", "").lower() or "dphil" in i.get("degree", "").lower()
        ]
        pdoc_advisees = [
            {
                "name": person.get("name", "missing name"),
                "type": "advisee",
                "advis_type": "postdoc",
                "interaction_date": get_dates(i).get("end_date", get_dates(i).get("date", dt.date.today())),
            }
            for i in relevant_emes
            if i.get("status") == "postdoc"
        ]
        if rc.postdoc_since_date:
            pdoc_advisees = [i for i in pdoc_advisees if rc.postdoc_since_date < i.get("interaction_date")]
        advisees.extend(phd_advisees)
        advisees.extend(pdoc_advisees)
    return retrieve_names_and_insts(rc, advisees)


def filter_since_date(pubs, rc):
    """Filter the publications after the since_date."""
    for pub in pubs:
        if isinstance(pub.get("year"), str):
            pub["year"] = int(pub.get("year"))
        pub["day"] = int(pub.get("day", 28))
        pub["month"] = pub.get("month", "dec")
        if pub.get("month").casefold().strip() == "tbd" or pub.get("month").strip() == "":
            print("WARNING: {} is missing month".format(pub["_id"]))
            pub["month"] = "dec"
        if is_after(pub, rc.pub_since_date):
            yield pub


def get_since_dates(rc):
    """Get the since_date from command."""
    if isinstance(rc.date, str):
        rc.pub_since_date = dt.datetime.strptime(rc, "%Y-%m-%d").date() - relativedelta(months=NUM_COAUTHOR_MONTHS)
    else:
        rc.pub_since_date = rc.date - relativedelta(months=NUM_COAUTHOR_MONTHS)
    if NUM_POSTDOC_MONTHS:
        if isinstance(rc, str):
            rc.postdoc_since_date = dt.datetime.strptime(rc.date, "%Y-%m-%d").date() - relativedelta(
                months=NUM_POSTDOC_MONTHS
            )
        else:
            rc.pub_since_date = rc.date - relativedelta(months=NUM_COAUTHOR_MONTHS)
    else:
        rc.postdoc_since_date = None
    return


def get_coauthors_from_pubs(rc, pubs, not_person):
    """Get co-authors' names from the publication. Not include the person itself."""
    not_person_akas = [not_person["_id"], not_person["name"]] + not_person["aka"]
    my_collabs = list()
    for pub in pubs:
        pub_date = dt.date(int(pub.get("year")), month_to_int(pub.get("month")), 1)
        my_collabs.extend(
            [
                {"name": collabs, "interaction_date": pub_date, "type": "co-author"}
                for collabs in (names for names in pub.get("author", []))
            ]
        )
    my_collabs.sort(key=lambda x: x["interaction_date"], reverse=True)
    coauthors = retrieve_names_and_insts(rc, my_collabs, not_person_akas)
    return coauthors


def retrieve_names_and_insts(rc, collabs, not_person_akas=[]):
    collab_buffer, my_collab_set = [], []
    for collab in collabs:
        person = fuzzy_retrieval(
            all_docs_from_collection(rc.client, "people"),
            ["name", "aka", "_id"],
            collab["name"],
            case_sensitive=False,
        )
        if not person:
            person = fuzzy_retrieval(
                all_docs_from_collection(rc.client, "contacts"),
                ["name", "aka", "_id"],
                collab["name"],
                case_sensitive=False,
            )
            if not person:
                if collab["name"] == "missing name":
                    print(
                        f"WARNING: a {collab.get('advis_type')} appointment "
                        f"was found for the target {collab.get('type')} but "
                        f"no name was specified. Please add an 'advisor' field "
                        f"for that education/employment entry in the database."
                    )
                else:
                    print(f"WARNING: {collab['name']} not found in contacts or people.")
                person = {"_id": collab["name"], "name": collab["name"], "type": collab.get("type")}
        if person.get("name", ""):
            collab["name"] = HumanName(person.get("name", ""))
        else:
            print("missing_person", person)
        collab["_id"] = person.get("_id")
        pinst = get_recent_org(person)
        inst = fuzzy_retrieval(
            all_docs_from_collection(rc.client, "institutions"),
            ["name", "aka", "_id"],
            pinst,
            case_sensitive=False,
        )
        if inst:
            collab["institution"] = inst["name"]
        else:
            collab["institution"] = pinst
            print(f"WARNING: {pinst} for {person.get('_id')} missing from institutions")
        if collab["_id"] not in collab_buffer and collab["name"] not in not_person_akas:
            my_collab_set.append(collab)
            collab_buffer.append(collab["_id"])
    return my_collab_set


def get_recent_org(person_info):
    """Get the person's most recent organization."""
    if "institution" in person_info:
        organization = person_info.get("institution")
    elif "employment" in person_info:
        employment = person_info.get("employment", []) + person_info.get("education", [])
        if len(employment) == 0:
            return ""
        # sort by end_year
        employment = sorted(employment, key=lambda d: d.get("end_year", float("inf")), reverse=True)
        organization = employment[0].get("organization", employment[0].get("institution"))
    else:
        organization = ""
    return organization


def query_people_and_institutions(rc, names):
    """Get the people and institutions names."""
    people, institutions, latest_active = [], [], []
    for person_name in names:
        person_found = fuzzy_retrieval(
            all_docs_from_collection(rc.client, "people"),
            ["name", "aka", "_id"],
            person_name[0],
            case_sensitive=False,
        )
        if not person_found:
            person_found = fuzzy_retrieval(
                all_docs_from_collection(rc.client, "contacts"),
                ["name", "aka", "_id"],
                person_name[0],
                case_sensitive=False,
            )
            if not person_found:
                print(
                    "WARNING: {} not found in contacts or people. Check aka".format(person_name[0]).encode("utf-8")
                )
            else:
                people.append(person_found["name"])
                inst = fuzzy_retrieval(
                    all_docs_from_collection(rc.client, "institutions"),
                    ["name", "aka", "_id"],
                    person_found["institution"],
                    case_sensitive=False,
                )
                if inst:
                    institutions.append(inst["name"])
                else:
                    institutions.append(person_found.get("institution", "missing"))
                    print("WARNING: {} missing from institutions".format(person_found["institution"]))
        else:
            people.append(person_found["name"])
            pinst = get_recent_org(person_found)
            inst = fuzzy_retrieval(
                all_docs_from_collection(rc.client, "institutions"),
                ["name", "aka", "_id"],
                pinst,
                case_sensitive=False,
            )
            if inst:
                institutions.append(inst["name"])
            else:
                institutions.append(pinst)
                print("WARNING: {} missing from institutions".format(pinst))
        latest_active.append(person_name[1])
    return people, institutions, latest_active


def get_inst_name(person, rc):
    """Get the name of instituion of the person's latest employment."""
    if "employment" in person:
        org = get_recent_org(person)
        person_inst_abbr = org
    elif "institution" in person:
        person_inst_abbr = person.get("institution")
    else:
        person_inst_abbr = ""
    person_inst = fuzzy_retrieval(
        all_docs_from_collection(rc.client, "institutions"),
        ["name", "aka", "_id"],
        person_inst_abbr,
        case_sensitive=False,
    )
    if person_inst is not None:
        person_inst_name = person_inst.get("name")
    else:
        person_inst_name = person_inst_abbr
        print(f"WARNING: {person_inst_abbr} is not found in institutions.")
    return person_inst_name


def get_person_pubs(coll, person):
    """Get the publications from one person."""
    my_names = frozenset(person.get("aka", []) + [person["name"]])
    pubs = filter_publications(coll, my_names, reverse=True, bold=False)
    return pubs


def make_person_4tups(person, rc):
    if "name" not in person:
        print("Warning")
    name = HumanName(person["name"])
    inst = get_inst_name(person, rc)
    first_names = " ".join([name.first, name.middle])
    return [(name.last, first_names, inst, "pi")]


def format_last_first_institution_names(rc, ppl_names, excluded_inst_name=None):
    """Get the last name, first name and institution name."""
    ppl = []
    for ppl_tup in ppl_names:
        inst = fuzzy_retrieval(
            all_docs_from_collection(rc.client, "institutions"),
            ["aka", "name", "_id"],
            ppl_tup[1],
            case_sensitive=False,
        )
        if inst:
            inst_name = inst.get("name", "")
        else:
            inst_name = ppl_tup[1]
        # remove all people who are in the institution of the person
        if inst_name != excluded_inst_name:
            name = HumanName(ppl_tup[0])
            yield name.last, " ".join([name.first, name.middle]), ppl_tup[1], " ", ppl_tup[2]
    return ppl


def format_to_nsf(tups, type_str):
    """Format the 3 tups to 2 tups. ('type_str', 'last, first', 'inst', ...)."""
    return [(type_str, "{}, {}".format(tup[0], tup[1])) + tup[2:] for tup in tups]


def apply_cell_style(*cells, style):
    """Apply the format to cells."""
    for cell in cells:
        cell.font = style["font"]
        cell.border = style["border"]
        cell.fill = style["fill"]
        cell.alignment = style["alignment"]
    return


def copy_cell_style(style_ref_cell):
    """Copy the cell format to a dictionary."""
    template_cell_style = {
        "font": copy(style_ref_cell.font),
        "border": copy(style_ref_cell.border),
        "fill": copy(style_ref_cell.fill),
        "alignment": copy(style_ref_cell.alignment),
    }
    return template_cell_style


def is_merged(cell):
    """Whether the cell is merged."""
    return True if type(cell).__name__ == "MergedCell" else False


def find_merged_cell(cells):
    """Find the index of all merged cells."""
    i, j = 0, 0
    lst = []
    while i < len(cells):
        if is_merged(cells[i]):
            while j < len(cells):
                if is_merged(cells[j]):
                    j += 1
                else:
                    i = j
                    break
            lst.append((i, j - 1))
        else:
            i += 1
    return lst


def unmerge(ws, cells):
    """Unmerge the cells."""
    lst = find_merged_cell(cells)
    for start, end in lst:
        ws.unmerge_cells("{}:{}").format(cells[start].coordinate, cells[end].coordinate)
    return


def get_person(person_id, rc):
    """Get the person's name."""
    person_found = fuzzy_retrieval(
        all_docs_from_collection(rc.client, "people"), ["name", "aka", "_id"], person_id, case_sensitive=False
    )
    if person_found:
        return person_found
    person_found = fuzzy_retrieval(
        all_docs_from_collection(rc.client, "contacts"), ["name", "aka", "_id"], person_id, case_sensitive=False
    )
    if not person_found:
        print("WARNING: {} missing from people and contacts. Check aka.".format(person_id))
        person_found = {"name": person_id}
    return person_found


def find_coeditors(person, rc):
    """Get the coeditors info of the person. Return (last, first, inst, journal)."""
    emps = person.get("employment")
    if emps is None:
        return set()

    def coeditor_id_journals(_emps):
        for emp in _emps:
            if emp.get("position") == "editor":
                _journal = emp.get("department", "")
                coeditor_ids = emp.get("coworkers", [])
                for _coeditor_id in coeditor_ids:
                    yield _coeditor_id, _journal

    coeditor_inst_journals = set()
    for coeditor_id, journal in coeditor_id_journals(emps):
        coeditor = get_person(coeditor_id, rc)
        coeditor_name = HumanName(coeditor.get("name"), "")
        inst_name = get_inst_name(coeditor, rc)
        coeditor_inst_journals.add((coeditor_name.last, coeditor_name.first, inst_name, journal, "co-editor"))
    return coeditor_inst_journals


class RecentCollaboratorsBuilder(BuilderBase):
    """Build recent collaborators from database entries"""

    btype = "recent-collabs"
    needed_colls = ["citations", "people", "contacts", "institutions"]

    def __init__(self, rc):
        """Initiate the class instance."""
        super().__init__(rc)
        self.template = os.path.join(
            os.path.dirname(os.path.dirname(__file__)), "templates", "coa_template_nsf.xlsx"
        )
        self.template2 = os.path.join(
            os.path.dirname(os.path.dirname(__file__)), "templates", "coa_template_doe.xlsx"
        )
        self.cmds = ["excel"]
        try:
            rc.verbose
        except AttributeError:
            rc.verbose = False

    def construct_global_ctx(self):
        """Construct the global ctx including database and methods."""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        gtx["people"] = sorted(
            all_docs_from_collection(rc.client, "people"),
            key=position_key,
            reverse=True,
        )
        gtx["contacts"] = sorted(
            all_docs_from_collection(rc.client, "contacts"),
            key=position_key,
            reverse=True,
        )
        gtx["institutions"] = all_docs_from_collection(rc.client, "institutions")
        gtx["citations"] = all_docs_from_collection(rc.client, "citations")
        gtx["all_docs_from_collection"] = all_docs_from_collection

    def query_ppl(self, target):
        """Query the data base for the target's collaborators' information."""
        rc = self.rc
        gtx = self.gtx
        person = fuzzy_retrieval(
            all_docs_from_collection(rc.client, "people"), ["aka", "name", "_id"], target, case_sensitive=False
        )
        if not person:
            raise RuntimeError("Person {} not found in people.".format(target).encode("utf-8"))

        pubs = get_person_pubs(gtx["citations"], person)
        pubs = filter_since_date(pubs, rc)
        try:
            if rc.verbose:
                for pub in pubs:
                    print(f"{pub.get('title')}, ({pub.get('year')})")
        except AttributeError:
            pass
        my_collabs = get_coauthors_from_pubs(rc, pubs, person)
        advisors = get_advisors_name_inst(person, rc)
        advisees = get_advisees_name_inst(all_docs_from_collection(rc.client, "people"), person, rc)
        collabs = []
        adviseors = advisors + advisees
        for collab in my_collabs:
            col_bool = True
            for advis in adviseors:
                if (
                    collab.get("name").last == advis.get("name").last
                    and collab.get("name").first == advis.get("name").first
                ):
                    col_bool = False
                    if advis.get("interaction_date"):
                        if collab.get("interaction_date", dt.date.today()) > advis.get(
                            "interaction_date", dt.date.today()
                        ):
                            advis.update({"interaction_date": collab.get("interaction_date")})
                    try:
                        if collab.get("interaction_date") > advis.get("interaction_date"):
                            advis.update({"interaction_date": collab.get("interaction_date")})
                    except TypeError:
                        print(f"ERROR: incorrect dates for an education/employment in {collab.get('name')}")
                        print(
                            f"collab date: {collab.get('interaction_date')}, "
                            f"advisee date: {advis.get('interaction_date')}"
                        )
                        raise
            if col_bool:
                collabs.append(collab)
        collabs.extend(advisees)
        collabs.extend(advisors)
        collabs.sort(key=lambda d: d["name"].last)
        if rc.verbose:
            output = [
                f"{my_collab.get('name').last}, "
                f"{my_collab.get('name').first}, "
                f"{my_collab.get('institution')}, "
                f"{my_collab.get('interaction_date')}, "
                f"{my_collab.get('advis_type', '')}, "
                f"{my_collab.get('type')}\n"
                for my_collab in collabs
            ]
            print(*output)
        person["name"] = HumanName(person.get("name"))
        results = {"person_info": person, "collabs": collabs}
        return results

    @staticmethod
    def fill_in_tab(ws, ppl, start_row, template_cell_style=None, cols="ABCDE"):
        """Add the information in person, institution pairs into the table 4 in nsf table."""
        nsf_mappings = {
            "co-author": "A:",
            "collaborator": "C:",
            "phd_advisor": "G:",
            "phd_advisee": "T:",
            "co-editor": "E:",
        }
        more_rows = len(ppl) - 1
        if more_rows > 0:
            ws.insert_rows(start_row, amount=more_rows)
        for row, person in enumerate(ppl, start=start_row):
            cells = [ws[f"{col}{row}"] for col in cols]
            if template_cell_style is not None:
                apply_cell_style(*cells, style=template_cell_style)
                ws[f"A{row}"].value = nsf_mappings.get(person.get("type"))
                ws[f"B{row}"].value = f"{person.get('name').last}, " f"{person.get('name').first}"
                ws[f"C{row}"].value = person.get("institution")
                if isinstance(person.get("interaction_date"), dt.date):
                    ws[f"E{row}"].value = person.get("interaction_date", dt.date.today()).strftime("%m/%d/%Y")
        return

    def render_template1(self, person_info, collabs, **kwargs):
        """Render the nsf template."""
        template = self.template
        wb = openpyxl.load_workbook(template)
        ws = wb.worksheets[0]
        nsf_collabs = copy(collabs)
        advis, coauths, coeditors = [], [], []
        pi_row_number = 17
        ws[f"B{pi_row_number}"].value = f"{person_info.get('name').last}, " f"{person_info.get('name').first}"
        ws[f"C{pi_row_number}"].value = person_info.get("institution")
        ws[f"D{pi_row_number}"].value = person_info.get("interaction_date", dt.date.today()).strftime("%m/%d/%Y")

        for collab in nsf_collabs:
            if collab.get("type") == "advisor" and collab.get("advis_type") == "phd":
                collab.update({"type": "phd_advisor"})
                advis.append(collab)
            if collab.get("type") == "advisee" and collab.get("advis_type") == "phd":
                collab.update({"type": "phd_advisee"})
                advis.append(collab)
            if collab.get("type") == "co-author":
                coauths.append(collab)
            if collab.get("type") == "co-editor":
                coeditors.append(collab)
        style = copy_cell_style(ws["A17"])

        if coauths:
            self.fill_in_tab(ws, coauths, start_row=52, template_cell_style=style)
        if advis:
            self.fill_in_tab(ws, advis, start_row=38, template_cell_style=style)
        # if person_info:
        #     self.fill_in_tab(
        #         ws, [person_info], start_row=17, template_cell_style=style
        #     )
        if coeditors:
            self.fill_in_tab(ws, coeditors, start_row=120, template_cell_style=style)
        wb.save(os.path.join(self.bldir, "{}_nsf.xlsx".format(person_info["_id"])))
        return locals()

    def render_template2(self, person_info, collabs, **kwargs):
        """Render the doe template."""

        template2 = self.template2
        num_rows = len(collabs)
        wb = openpyxl.load_workbook(template2)
        ws = wb.worksheets[0]
        for row in range(num_rows):
            ws["A{}".format(row + 8)].value = person_info["name"].last
            ws["B{}".format(row + 8)].value = person_info["name"].first
            ws["C{}".format(row + 8)].value = collabs[row]["name"].last
            ws["D{}".format(row + 8)].value = collabs[row]["name"].first
            ws["F{}".format((row + 8))].value = collabs[row]["institution"]
            ws["G{}".format((row + 8))].value = collabs[row]["type"]
            if isinstance(collabs[row]["interaction_date"], dt.date):
                ws["H{}".format((row + 8))].value = collabs[row]["interaction_date"].year
        ws["C3"].value = person_info["name"].full_name
        ws["C4"].value = person_info.get("email", "")
        wb.save(os.path.join(self.bldir, "{}_doe.xlsx".format(person_info["_id"])))
        return locals()

    def excel(self):
        """Query data base and build nsf and doe excels."""
        rc = self.rc
        if not rc.people:
            sys.exit("please rerun specifying --people PERSON")
        if isinstance(rc.people, str):
            rc.people = [rc.people]
        if not hasattr(rc, "date"):
            rc.date = dt.date.today()
        get_since_dates(rc)
        rc.post_doc_window_months = NUM_POSTDOC_MONTHS
        print(f"filtering coauthors for papers since {rc.pub_since_date}")
        for target in rc.people:
            query_results = self.query_ppl(target)
            self.render_template1(**query_results)
            self.render_template2(**query_results)
